from infrahub_sdk import InfrahubClient
from infrahub_sdk.protocols import (
    CoreArtifact,
    CoreArtifactDefinition,
    CoreCheckDefinition,
    CoreRepository,
    CoreRepositoryValidator,
    CoreUserValidator,
)
from infrahub_sdk.uuidt import UUIDT
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub import lock
from infrahub.context import InfrahubContext
from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus, ValidatorConclusion
from infrahub.core.registry import registry
from infrahub.exceptions import CheckError, RepositoryError
from infrahub.message_bus import Meta, messages
from infrahub.services import InfrahubServices
from infrahub.validators.tasks import start_validator
from infrahub.worker import WORKER_IDENTITY

from ..core.manager import NodeManager
from ..core.timestamp import Timestamp
from ..core.validators.checks_runner import run_checks_and_update_validator
from ..log import get_log_data
from ..tasks.artifact import define_artifact
from ..workflows.catalogue import (
    GIT_REPOSITORY_MERGE_CONFLICTS_CHECKS_RUN,
    GIT_REPOSITORY_USER_CHECK_RUN,
    GIT_REPOSITORY_USER_CHECKS_DEFINITIONS_TRIGGER,
    REQUEST_ARTIFACT_DEFINITION_GENERATE,
    REQUEST_ARTIFACT_GENERATE,
)
from ..workflows.utils import add_branch_tag, add_tags
from .models import (
    CheckRepositoryMergeConflicts,
    GitDiffNamesOnly,
    GitDiffNamesOnlyResponse,
    GitRepositoryAdd,
    GitRepositoryAddReadOnly,
    GitRepositoryImportObjects,
    GitRepositoryMerge,
    GitRepositoryPullReadOnly,
    RequestArtifactDefinitionGenerate,
    RequestArtifactGenerate,
    TriggerRepositoryInternalChecks,
    TriggerRepositoryUserChecks,
    UserCheckData,
    UserCheckDefinitionData,
)
from .repository import InfrahubReadOnlyRepository, InfrahubRepository, get_initialized_repo


@flow(
    name="git-repository-add-read-write",
    flow_run_name="Adding repository {model.repository_name} in branch {model.infrahub_branch_name}",
)
async def add_git_repository(model: GitRepositoryAdd, service: InfrahubServices) -> None:
    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])

    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        repo = await InfrahubRepository.new(
            id=model.repository_id,
            name=model.repository_name,
            location=model.location,
            client=service.client,
            infrahub_branch_name=model.infrahub_branch_name,
            internal_status=model.internal_status,
            default_branch_name=model.default_branch_name,
            service=service,
        )
        await repo.import_objects_from_files(  # type: ignore[call-overload]
            infrahub_branch_name=model.infrahub_branch_name, git_branch_name=model.default_branch_name
        )
        if model.internal_status == RepositoryInternalStatus.ACTIVE.value:
            await repo.sync()

            # Notify other workers they need to clone the repository
            notification = messages.RefreshGitFetch(
                meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                location=model.location,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                repository_kind=InfrahubKind.REPOSITORY,
                infrahub_branch_name=model.infrahub_branch_name,
                infrahub_branch_id=model.infrahub_branch_id,
            )
            await service.message_bus.send(message=notification)


@flow(
    name="git-repository-add-read-only",
    flow_run_name="Adding read only repository {model.repository_name} in branch {model.infrahub_branch_name}",
)
async def add_git_repository_read_only(model: GitRepositoryAddReadOnly, service: InfrahubServices) -> None:
    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])

    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        repo = await InfrahubReadOnlyRepository.new(
            id=model.repository_id,
            name=model.repository_name,
            location=model.location,
            client=service.client,
            ref=model.ref,
            infrahub_branch_name=model.infrahub_branch_name,
            service=service,
        )
        await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name)  # type: ignore[call-overload]
        if model.internal_status == RepositoryInternalStatus.ACTIVE.value:
            await repo.sync_from_remote()

            # Notify other workers they need to clone the repository
            notification = messages.RefreshGitFetch(
                meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                location=model.location,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                repository_kind=InfrahubKind.REPOSITORY,
                infrahub_branch_name=model.infrahub_branch_name,
                infrahub_branch_id=model.infrahub_branch_id,
            )
            await service.message_bus.send(message=notification)


@flow(name="git-repositories-create-branch", flow_run_name="Create branch '{branch}' in Git Repositories")
async def create_branch(branch: str, branch_id: str, service: InfrahubServices) -> None:
    """Request to the creation of git branches in available repositories."""
    await add_tags(branches=[branch])
    repositories: list[CoreRepository] = await service.client.filters(kind=CoreRepository)
    batch = await service.client.create_batch()
    for repository in repositories:
        batch.add(
            task=git_branch_create,
            client=service.client.client,
            branch=branch,
            branch_id=branch_id,
            repository_name=repository.name.value,
            repository_id=repository.id,
            repository_location=repository.location.value,
            service=service,
        )

    async for _, _ in batch.execute():
        pass


@flow(name="git_repositories_sync", flow_run_name="Sync Git Repositories")
async def sync_remote_repositories(service: InfrahubServices) -> None:
    log = get_run_logger()

    branches = await service.client.branch.all()
    repositories = await service.client.get_list_repositories(branches=branches, kind=InfrahubKind.REPOSITORY)

    for repo_name, repository_data in repositories.items():
        active_internal_status = RepositoryInternalStatus.ACTIVE.value
        default_internal_status = repository_data.branch_info[registry.default_branch].internal_status
        staging_branch = None
        if default_internal_status != RepositoryInternalStatus.ACTIVE.value:
            active_internal_status = RepositoryInternalStatus.STAGING.value
            staging_branch = repository_data.get_staging_branch()

        infrahub_branch = staging_branch or registry.default_branch

        async with lock.registry.get(name=repo_name, namespace="repository"):
            init_failed = False
            try:
                repo = await InfrahubRepository.init(
                    service=service,
                    id=repository_data.repository.id,
                    name=repository_data.repository.name.value,
                    location=repository_data.repository.location.value,
                    client=service.client,
                    internal_status=active_internal_status,
                    default_branch_name=repository_data.repository.default_branch.value,
                )
            except RepositoryError as exc:
                service.log.error(str(exc))
                init_failed = True

            if init_failed:
                try:
                    repo = await InfrahubRepository.new(
                        service=service,
                        id=repository_data.repository.id,
                        name=repository_data.repository.name.value,
                        location=repository_data.repository.location.value,
                        client=service.client,
                        internal_status=active_internal_status,
                        default_branch_name=repository_data.repository.default_branch.value,
                    )
                    await repo.import_objects_from_files(  # type: ignore[call-overload]
                        git_branch_name=registry.default_branch, infrahub_branch_name=infrahub_branch
                    )
                except RepositoryError as exc:
                    log.info(exc.message)
                    continue

            try:
                await repo.sync(staging_branch=staging_branch)
                # Tell workers to fetch to stay in sync
                message = messages.RefreshGitFetch(
                    meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                    location=repository_data.repository.location.value,
                    repository_id=repository_data.repository.id,
                    repository_name=repository_data.repository.name.value,
                    repository_kind=repository_data.repository.get_kind(),
                    infrahub_branch_name=infrahub_branch,
                    infrahub_branch_id=branches[infrahub_branch].id,
                )
                await service.message_bus.send(message=message)
            except RepositoryError as exc:
                log.info(exc.message)


@task(  # type: ignore[arg-type]
    name="git-branch-create",
    task_run_name="Create branch '{branch}' in repository {repository_name}",
    cache_policy=NONE,
)
async def git_branch_create(
    client: InfrahubClient,
    branch: str,
    branch_id: str,
    repository_id: str,
    repository_name: str,
    repository_location: str,
    service: InfrahubServices,
) -> None:
    log = get_run_logger()
    repo = await InfrahubRepository.init(
        id=repository_id, name=repository_name, location=repository_location, client=client, service=service
    )

    async with lock.registry.get(name=repository_name, namespace="repository"):
        await repo.create_branch_in_git(branch_name=branch, branch_id=branch_id, push_origin=True)

        # New branch has been pushed remotely, tell workers to fetch it
        message = messages.RefreshGitFetch(
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
            location=repo.get_location(),
            repository_id=str(repo.id),
            repository_name=repo.name,
            repository_kind=InfrahubKind.REPOSITORY,
            infrahub_branch_name=branch,
            infrahub_branch_id=branch_id,
        )
        await service.message_bus.send(message=message)
        log.debug("Sent message to all workers to fetch the latest version of the repository (RefreshGitFetch)")


@flow(name="artifact-definition-generate", flow_run_name="Generate all artifacts")
async def generate_artifact_definition(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_branch_tag(branch_name=branch)

    artifact_definitions = await service.client.all(kind=CoreArtifactDefinition, branch=branch, include=["id"])

    for artifact_definition in artifact_definitions:
        model = RequestArtifactDefinitionGenerate(
            branch=branch,
            artifact_definition_id=artifact_definition.id,
            artifact_definition_name=artifact_definition.name.value,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_ARTIFACT_DEFINITION_GENERATE, context=context, parameters={"model": model}
        )


@flow(name="artifact-generate", flow_run_name="Generate artifact {model.artifact_name}")
async def generate_artifact(model: RequestArtifactGenerate, service: InfrahubServices) -> None:
    await add_tags(branches=[model.branch_name], nodes=[model.target_id])
    log = get_run_logger()
    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=service,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )

    artifact, artifact_created = await define_artifact(model=model, service=service)

    try:
        result = await repo.render_artifact(artifact=artifact, artifact_created=artifact_created, message=model)
        log.debug(
            f"Generated artifact | changed: {result.changed} | {result.checksum} | {result.storage_id}",
        )
    except Exception:
        log.exception("Failed to generate artifact")
        artifact.status.value = "Error"
        await artifact.save()
        raise


@flow(
    name="request_artifact_definitions_generate",
    flow_run_name="Trigger Generation of Artifacts for {model.artifact_definition_name}",
)
async def generate_request_artifact_definition(
    model: RequestArtifactDefinitionGenerate, context: InfrahubContext, service: InfrahubServices
) -> None:
    await add_tags(branches=[model.branch])

    artifact_definition = await service.client.get(
        kind=CoreArtifactDefinition, id=model.artifact_definition_id, branch=model.branch
    )

    await artifact_definition.targets.fetch()
    group = artifact_definition.targets.peer
    await group.members.fetch()
    current_members = [member.id for member in group.members.peers]

    existing_artifacts = await service.client.filters(
        kind=CoreArtifact,
        definition__ids=[model.artifact_definition_id],
        include=["object"],
        branch=model.branch,
    )
    artifacts_by_member = {}
    for artifact in existing_artifacts:
        if artifact.object.id in current_members:
            artifacts_by_member[artifact.object.peer.id] = artifact.id

    await artifact_definition.transformation.fetch()
    transformation_repository = artifact_definition.transformation.peer.repository

    await transformation_repository.fetch()

    transform = artifact_definition.transformation.peer
    await transform.query.fetch()
    query = transform.query.peer
    repository = transformation_repository.peer
    branch = await service.client.branch.get(branch_name=model.branch)
    if branch.sync_with_git:
        repository = await service.client.get(
            kind=InfrahubKind.GENERICREPOSITORY, id=repository.id, branch=model.branch, fragment=True
        )
    transform_location = ""

    if transform.typename == InfrahubKind.TRANSFORMJINJA2:
        transform_location = transform.template_path.value
    elif transform.typename == InfrahubKind.TRANSFORMPYTHON:
        transform_location = f"{transform.file_path.value}::{transform.class_name.value}"

    for relationship in group.members.peers:
        member = relationship.peer
        artifact_id = artifacts_by_member.get(member.id)
        if model.limit and artifact_id not in model.limit:
            continue

        request_artifact_generate_model = RequestArtifactGenerate(
            artifact_name=artifact_definition.artifact_name.value,
            artifact_id=artifact_id,
            artifact_definition=model.artifact_definition_id,
            commit=repository.commit.value,
            content_type=artifact_definition.content_type.value,
            transform_type=str(transform.typename),
            transform_location=transform_location,
            repository_id=repository.id,
            repository_name=repository.name.value,
            repository_kind=repository.get_kind(),
            branch_name=model.branch,
            query=query.name.value,
            variables=member.extract(params=artifact_definition.parameters.value),
            target_id=member.id,
            target_name=member.display_label,
            target_kind=member.get_kind(),
            timeout=transform.timeout.value,
            context=context,
        )

        await service.workflow.submit_workflow(
            workflow=REQUEST_ARTIFACT_GENERATE, context=context, parameters={"model": request_artifact_generate_model}
        )


@flow(name="git-repository-pull-read-only", flow_run_name="Pull latest commit on {model.repository_name}")
async def pull_read_only(model: GitRepositoryPullReadOnly, service: InfrahubServices) -> None:
    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])
    log = get_run_logger()

    if not model.ref and not model.commit:
        log.warning("No commit or ref in GitRepositoryPullReadOnly message")
        return
    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        init_failed = False
        try:
            repo = await InfrahubReadOnlyRepository.init(
                id=model.repository_id,
                name=model.repository_name,
                location=model.location,
                client=service.client,
                ref=model.ref,
                infrahub_branch_name=model.infrahub_branch_name,
                service=service,
            )
        except RepositoryError:
            init_failed = True

        if init_failed:
            repo = await InfrahubReadOnlyRepository.new(
                id=model.repository_id,
                name=model.repository_name,
                location=model.location,
                client=service.client,
                ref=model.ref,
                infrahub_branch_name=model.infrahub_branch_name,
                service=service,
            )

        await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name, commit=model.commit)  # type: ignore[call-overload]
        await repo.sync_from_remote(commit=model.commit)

        # Tell workers to fetch to stay in sync
        message = messages.RefreshGitFetch(
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
            location=model.location,
            repository_id=model.repository_id,
            repository_name=model.repository_name,
            repository_kind=InfrahubKind.READONLYREPOSITORY,
            infrahub_branch_name=model.infrahub_branch_name,
            infrahub_branch_id=model.infrahub_branch_id,
        )
        await service.message_bus.send(message=message)


@flow(
    name="git-repository-merge",
    flow_run_name="Merge {model.source_branch} > {model.destination_branch} in git repository",
)
async def merge_git_repository(model: GitRepositoryMerge, service: InfrahubServices) -> None:
    await add_tags(branches=[model.source_branch, model.destination_branch], nodes=[model.repository_id])

    repo = await InfrahubRepository.init(
        id=model.repository_id,
        name=model.repository_name,
        client=service.client,
        default_branch_name=model.default_branch,
        service=service,
    )

    if model.internal_status == RepositoryInternalStatus.STAGING.value:
        repo_source = await service.client.get(
            kind=InfrahubKind.GENERICREPOSITORY, id=model.repository_id, branch=model.source_branch
        )
        repo_main = await service.client.get(kind=InfrahubKind.GENERICREPOSITORY, id=model.repository_id)
        repo_main.internal_status.value = RepositoryInternalStatus.ACTIVE.value
        repo_main.sync_status.value = repo_source.sync_status.value

        commit = repo.get_commit_value(branch_name=repo.default_branch, remote=False)
        repo_main.commit.value = commit

        await repo_main.save()
    else:
        async with lock.registry.get(name=model.repository_name, namespace="repository"):
            await repo.merge(source_branch=model.source_branch, dest_branch=model.destination_branch)
            if repo.location:
                # Destination branch has changed and pushed remotely, tell workers to re-fetch
                message = messages.RefreshGitFetch(
                    meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                    location=repo.location,
                    repository_id=str(repo.id),
                    repository_name=repo.name,
                    repository_kind=InfrahubKind.REPOSITORY,
                    infrahub_branch_name=model.destination_branch,
                    infrahub_branch_id=model.destination_branch_id,
                )
                await service.message_bus.send(message=message)


@flow(name="git-repository-import-object", flow_run_name="Import objects from git repository")
async def import_objects_from_git_repository(model: GitRepositoryImportObjects, service: InfrahubServices) -> None:
    await add_branch_tag(model.infrahub_branch_name)
    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=service,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )
    await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name, commit=model.commit)  # type: ignore[call-overload]


@flow(
    name="git-repository-diff-names-only",
    flow_run_name="Collecting modifications between commits {model.first_commit} and {model.second_commit}",
    persist_result=True,
)
async def git_repository_diff_names_only(
    model: GitDiffNamesOnly, service: InfrahubServices
) -> GitDiffNamesOnlyResponse:
    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=service,
        repository_kind=model.repository_kind,
    )
    files_changed: list[str] = []
    files_removed: list[str] = []

    if model.second_commit:
        files_changed, files_added, files_removed = await repo.calculate_diff_between_commits(
            first_commit=model.first_commit, second_commit=model.second_commit
        )
    else:
        files_added = await repo.list_all_files(commit=model.first_commit)

    response = GitDiffNamesOnlyResponse(
        files_added=files_added, files_changed=files_changed, files_removed=files_removed
    )
    return response


@flow(
    name="git-repository-user-checks-definition-trigger",
    flow_run_name="Trigger user defined checks for repository {model.repository_name}",
)
async def trigger_repository_user_checks_definitions(
    model: UserCheckDefinitionData, context: InfrahubContext, service: InfrahubServices
) -> None:
    await add_tags(branches=[model.branch_name], nodes=[model.proposed_change])
    log = get_run_logger()

    definition = await service.client.get(
        kind=CoreCheckDefinition, id=model.check_definition_id, branch=model.branch_name
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)
    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []
    await proposed_change.validations.fetch()

    previous_validator: CoreUserValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer

        if (
            existing_validator.typename == InfrahubKind.USERVALIDATOR
            and existing_validator.repository.id == model.repository_id
            and existing_validator.check_definition.id == model.check_definition_id
        ):
            previous_validator = existing_validator
            service.log.info("Found the same validator", validator=previous_validator)

    validator = await start_validator(
        service=service,
        validator=previous_validator,
        validator_type=CoreUserValidator,
        proposed_change=model.proposed_change,
        data={
            "label": f"Check: {definition.name.value}",
            "repository": model.repository_id,
            "check_definition": model.check_definition_id,
        },
        context=context,
    )

    if definition.targets.id:
        # Check against a group of targets
        await definition.targets.fetch()
        group = definition.targets.peer
        await group.members.fetch()
        check_models = []
        for relationship in group.members.peers:
            member = relationship.peer

            check_execution_id = str(UUIDT())
            check_execution_ids.append(check_execution_id)
            check_model = UserCheckData(
                name=member.display_label,
                validator_id=validator.id,
                validator_execution_id=validator_execution_id,
                check_execution_id=check_execution_id,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                commit=model.commit,
                file_path=model.file_path,
                class_name=model.class_name,
                branch_name=model.branch_name,
                check_definition_id=model.check_definition_id,
                proposed_change=model.proposed_change,
                variables=member.extract(params=definition.parameters.value),
                branch_diff=model.branch_diff,
                timeout=definition.timeout.value,
            )
            check_models.append(check_model)
    else:
        check_execution_id = str(UUIDT())
        check_execution_ids.append(check_execution_id)
        check_models = [
            UserCheckData(
                name=definition.name.value,
                validator_id=validator.id,
                validator_execution_id=validator_execution_id,
                check_execution_id=check_execution_id,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                commit=model.commit,
                file_path=model.file_path,
                class_name=model.class_name,
                branch_name=model.branch_name,
                check_definition_id=model.check_definition_id,
                proposed_change=model.proposed_change,
                branch_diff=model.branch_diff,
                timeout=definition.timeout.value,
            )
        ]

    checks_in_execution = ",".join(check_execution_ids)
    log.info(f"Checks in execution {checks_in_execution}")

    checks_coroutines = [
        service.workflow.execute_workflow(
            workflow=GIT_REPOSITORY_USER_CHECK_RUN, parameters={"model": model}, expected_return=ValidatorConclusion
        )
        for model in check_models
    ]

    await run_checks_and_update_validator(
        checks=checks_coroutines,
        validator=validator,
        context=context,
        service=service,
        proposed_change_id=model.proposed_change,
    )


@flow(
    name="git-repository-trigger-user-checks",
    flow_run_name="Evaluating user-defined checks on repository {model.repository_name}",
)
async def trigger_user_checks(
    model: TriggerRepositoryUserChecks, service: InfrahubServices, context: InfrahubContext
) -> None:
    """Request to start validation checks on a specific repository for User-defined checks."""

    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])
    log = get_run_logger()

    repository = await service.client.get(
        kind=InfrahubKind.GENERICREPOSITORY, id=model.repository_id, branch=model.source_branch, fragment=True
    )
    await repository.checks.fetch()

    for relationship in repository.checks.peers:
        log.info("Adding check for user defined check")
        check_definition = relationship.peer
        user_check_definition_model = UserCheckDefinitionData(
            check_definition_id=check_definition.id,
            repository_id=repository.id,
            repository_name=repository.name.value,
            commit=repository.commit.value,
            file_path=check_definition.file_path.value,
            class_name=check_definition.class_name.value,
            branch_name=model.source_branch,
            proposed_change=model.proposed_change,
            branch_diff=model.branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=GIT_REPOSITORY_USER_CHECKS_DEFINITIONS_TRIGGER,
            context=context,
            parameters={"model": user_check_definition_model},
        )


@flow(
    name="git-repository-trigger-internal-checks",
    flow_run_name="Running repository checks for repository {model.repository}",
)
async def trigger_internal_checks(
    model: TriggerRepositoryInternalChecks, service: InfrahubServices, context: InfrahubContext
) -> None:
    """Request to start validation checks on a specific repository."""
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])
    log = get_run_logger()

    repository = await service.client.get(
        kind=InfrahubKind.GENERICREPOSITORY, id=model.repository, branch=model.source_branch
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []
    await proposed_change.validations.fetch()
    await repository.checks.fetch()

    validator_name = f"Repository Validator: {repository.name.value}"
    previous_validator: CoreRepositoryValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer

        if (
            existing_validator.typename == InfrahubKind.REPOSITORYVALIDATOR
            and existing_validator.repository.id == model.repository
        ):
            previous_validator = existing_validator

    validator = await start_validator(
        service=service,
        validator=previous_validator,
        validator_type=CoreRepositoryValidator,
        proposed_change=model.proposed_change,
        data={
            "label": validator_name,
            "repository": model.repository,
        },
        context=context,
    )

    check_execution_id = str(UUIDT())
    check_execution_ids.append(check_execution_id)
    log.info("Adding check for merge conflict")
    checks_in_execution = ",".join(check_execution_ids)
    log.info(f"Checks in execution {checks_in_execution}")

    check_merge_conflict_model = CheckRepositoryMergeConflicts(
        validator_id=validator.id,
        validator_execution_id=validator_execution_id,
        check_execution_id=check_execution_id,
        proposed_change=model.proposed_change,
        repository_id=model.repository,
        repository_name=repository.name.value,
        source_branch=model.source_branch,
        target_branch=model.target_branch,
    )
    check_coroutine = service.workflow.execute_workflow(
        workflow=GIT_REPOSITORY_MERGE_CONFLICTS_CHECKS_RUN,
        parameters={"model": check_merge_conflict_model},
        expected_return=ValidatorConclusion,
    )

    await run_checks_and_update_validator(
        checks=[check_coroutine],
        validator=validator,
        context=context,
        service=service,
        proposed_change_id=model.proposed_change,
    )


@flow(
    name="git-repository-check-merge-conflict",
    flow_run_name="Check for merge conflicts between {model.source_branch} and {model.target_branch}",
)
async def run_check_merge_conflicts(
    model: CheckRepositoryMergeConflicts, service: InfrahubServices
) -> ValidatorConclusion:
    """Runs a check to see if there are merge conflicts between two branches."""
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    success_condition = "-"
    validator = await service.client.get(kind=InfrahubKind.REPOSITORYVALIDATOR, id=model.validator_id)
    await validator.checks.fetch()

    repo = await InfrahubRepository.init(id=model.repository_id, name=model.repository_name, service=service)
    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        conflicts = await repo.get_conflicts(source_branch=model.source_branch, dest_branch=model.target_branch)

    existing_checks = {}
    for relationship in validator.checks.peers:
        existing_check = relationship.peer
        if existing_check.typename == InfrahubKind.FILECHECK and existing_check.kind.value == "MergeConflictCheck":
            check_key = ""
            if existing_check.files.value:
                check_key = "".join(existing_check.files.value)
            check_key = f"-{check_key}"
            existing_checks[check_key] = existing_check

    if conflicts:
        validator_conclusion = ValidatorConclusion.FAILURE
        for conflict in conflicts:
            conflict_key = f"-{conflict}"
            if conflict_key in existing_checks:
                existing_checks[conflict_key].created_at.value = Timestamp().to_string()
                await existing_checks[conflict_key].save()
                existing_checks.pop(conflict_key)
            else:
                check = await service.client.create(
                    kind=InfrahubKind.FILECHECK,
                    data={
                        "name": conflict,
                        "origin": "ConflictCheck",
                        "kind": "MergeConflictCheck",
                        "validator": model.validator_id,
                        "created_at": Timestamp().to_string(),
                        "files": [conflict],
                        "conclusion": "failure",
                        "severity": "critical",
                    },
                )
                await check.save()

    elif success_condition in existing_checks:
        validator_conclusion = ValidatorConclusion.SUCCESS
        existing_checks[success_condition].created_at.value = Timestamp().to_string()
        await existing_checks[success_condition].save()
        existing_checks.pop(success_condition)

    else:
        validator_conclusion = ValidatorConclusion.SUCCESS
        check = await service.client.create(
            kind=InfrahubKind.FILECHECK,
            data={
                "name": "Merge Conflict Check",
                "origin": "ConflictCheck",
                "kind": "MergeConflictCheck",
                "validator": model.validator_id,
                "created_at": Timestamp().to_string(),
                "conclusion": validator_conclusion.value,
                "severity": "info",
            },
        )
        await check.save()

    async with service.database.start_session() as db:
        await NodeManager.delete(db=db, nodes=list(existing_checks.values()))

    return validator_conclusion


@flow(name="git-repository-run-user-check", flow_run_name="Execute user defined Check '{model.name}'")
async def run_user_check(model: UserCheckData, service: InfrahubServices) -> ValidatorConclusion:
    await add_tags(branches=[model.branch_name], nodes=[model.proposed_change])
    log = get_run_logger()

    validator = await service.client.get(kind=InfrahubKind.USERVALIDATOR, id=model.validator_id)
    await validator.checks.fetch()

    repo = await InfrahubRepository.init(
        id=model.repository_id, name=model.repository_name, commit=model.commit, service=service
    )
    conclusion = ValidatorConclusion.FAILURE
    severity = "critical"
    try:
        check_run = await repo.execute_python_check.with_options(timeout_seconds=model.timeout)(
            branch_name=model.branch_name,
            location=model.file_path,
            class_name=model.class_name,
            client=service.client,
            commit=model.commit,
            params=model.variables,
        )  # type: ignore[misc]
        if check_run.passed:
            conclusion = ValidatorConclusion.SUCCESS
            severity = "info"
            log.info("The check passed")
        else:
            log.warning("The check reported failures")
            for log_entry in check_run.log_entries:
                log.warning(log_entry)
        log_entries = check_run.log_entries
    except CheckError as exc:
        log.warning("The check failed to run")
        log.error(exc.message)
        log_entries = f"FATAL Error/n:{exc.message}"

    check = None
    for relationship in validator.checks.peers:
        existing_check = relationship.peer
        if (
            existing_check.typename == InfrahubKind.STANDARDCHECK
            and existing_check.kind.value == "CheckDefinition"
            and existing_check.name.value == model.name
        ):
            check = existing_check

    if check:
        check.created_at.value = Timestamp().to_string()
        check.message.value = log_entries
        check.conclusion.value = conclusion.value
        check.severity.value = severity
        await check.save()
    else:
        check = await service.client.create(
            kind=InfrahubKind.STANDARDCHECK,
            data={
                "name": model.name,
                "origin": model.repository_id,
                "kind": "CheckDefinition",
                "validator": model.validator_id,
                "created_at": Timestamp().to_string(),
                "message": log_entries,
                "conclusion": conclusion.value,
                "severity": severity,
            },
        )
        await check.save()

    return conclusion



from prefect import flow

from infrahub.message_bus import messages
from infrahub.message_bus.messages.send_echo_request import SendEchoRequestResponse, SendEchoRequestResponseData
from infrahub.services import InfrahubServices


@flow(name="echo-request")
async def request(message: messages.SendEchoRequest, service: InfrahubServices) -> None:
    service.log.info(f"Received message: {message.message}")
    if message.reply_requested:
        response = SendEchoRequestResponse(data=SendEchoRequestResponseData(response=f"Reply to: {message.message}"))
        await service.message_bus.reply_if_initiator_meta(message=response, initiator=message)



from copy import deepcopy
from dataclasses import dataclass, field, replace
from typing import Iterable

from infrahub.core.constants import NULL_VALUE, DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType

from .model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffs,
    EnrichedDiffSingleRelationship,
)


@dataclass
class NodePair:
    earlier: EnrichedDiffNode | None = field(default=None)
    later: EnrichedDiffNode | None = field(default=None)


class DiffCombiner:
    def __init__(self) -> None:
        # {child_uuid: (parent_uuid, parent_rel_name)}
        self._child_parent_uuid_map: dict[str, tuple[str, str]] = {}
        self._parent_node_uuids: set[str] = set()
        self._earlier_nodes_by_uuid: dict[str, EnrichedDiffNode] = {}
        self._later_nodes_by_uuid: dict[str, EnrichedDiffNode] = {}
        self._common_node_uuids: set[str] = set()
        self._diff_branch_name: str | None = None

    def _initialize(self, earlier_diff: EnrichedDiffRoot, later_diff: EnrichedDiffRoot) -> None:
        self._diff_branch_name = earlier_diff.diff_branch_name
        self._child_parent_uuid_map = {}
        self._earlier_nodes_by_uuid = {}
        self._later_nodes_by_uuid = {}
        self._common_node_uuids = set()
        # map the parent of each node (if it exists), preference to the later diff
        for diff_root in (earlier_diff, later_diff):
            for child_node in diff_root.nodes:
                for parent_rel in child_node.relationships:
                    for parent_node in parent_rel.nodes:
                        self._child_parent_uuid_map[child_node.uuid] = (parent_node.uuid, parent_rel.name)
        # UUIDs of all the parents, removing the stale parents from the earlier diff
        self._parent_node_uuids = {parent_tuple[0] for parent_tuple in self._child_parent_uuid_map.values()}
        self._earlier_nodes_by_uuid = {n.uuid: n for n in earlier_diff.nodes}
        self._later_nodes_by_uuid = {n.uuid: n for n in later_diff.nodes}
        self._common_node_uuids = set(self._earlier_nodes_by_uuid.keys()) & set(self._later_nodes_by_uuid.keys())

    @property
    def diff_branch_name(self) -> str:
        if not self._diff_branch_name:
            raise RuntimeError("DiffCombiner is not initialized")
        return self._diff_branch_name

    def _filter_nodes_to_keep(self, earlier_diff: EnrichedDiffRoot, later_diff: EnrichedDiffRoot) -> list[NodePair]:
        filtered_node_pairs: list[NodePair] = []
        for earlier_node in earlier_diff.nodes:
            later_node: EnrichedDiffNode | None = None
            if earlier_node.uuid in self._common_node_uuids:
                later_node = self._later_nodes_by_uuid[earlier_node.uuid]
            # this is an out-of-date parent
            if (
                earlier_node.action is DiffAction.UNCHANGED
                and (later_node is None or later_node.action is DiffAction.UNCHANGED)
                and earlier_node.uuid not in self._parent_node_uuids
            ):
                continue
            if later_node is None:
                filtered_node_pairs.append(NodePair(earlier=earlier_node))
                continue
            # if node was added and removed or vice-versa, remove it from the diff
            if {earlier_node.action, later_node.action} == {DiffAction.ADDED, DiffAction.REMOVED}:
                continue
            filtered_node_pairs.append(NodePair(earlier=earlier_node, later=later_node))
        for later_node in later_diff.nodes:
            # these have already been handled
            if later_node.uuid in self._common_node_uuids:
                continue
            filtered_node_pairs.append(NodePair(later=later_node))
        return filtered_node_pairs

    def _get_parent_relationship_name(self, node_id: str) -> str | None:
        if node_id not in self._child_parent_uuid_map:
            return None
        return self._child_parent_uuid_map[node_id][1]

    def _should_include(self, earlier: DiffAction, later: DiffAction) -> bool:
        actions = {earlier, later}
        if actions == {DiffAction.UNCHANGED}:
            return False
        if earlier is DiffAction.ADDED and later is DiffAction.REMOVED:
            return False
        return True

    def _combine_actions(self, earlier: DiffAction, later: DiffAction) -> DiffAction:
        actions = {earlier, later}
        if len(actions) == 1:
            return actions.pop()
        if DiffAction.UNCHANGED in actions:
            actual_action = actions - {DiffAction.UNCHANGED}
            return actual_action.pop()
        actions_map = {
            (DiffAction.ADDED, DiffAction.REMOVED): DiffAction.UPDATED,
            (DiffAction.ADDED, DiffAction.UPDATED): DiffAction.ADDED,
            (DiffAction.UPDATED, DiffAction.ADDED): DiffAction.UPDATED,
            (DiffAction.UPDATED, DiffAction.REMOVED): DiffAction.REMOVED,
            (DiffAction.REMOVED, DiffAction.ADDED): DiffAction.UPDATED,
            (DiffAction.REMOVED, DiffAction.UPDATED): DiffAction.UPDATED,
        }
        return actions_map[earlier, later]

    @staticmethod
    def combine_conflicts(
        earlier: EnrichedDiffConflict | None, later: EnrichedDiffConflict | None
    ) -> EnrichedDiffConflict | None:
        if later is None:
            return None
        if earlier is None:
            return deepcopy(later)
        combined = replace(later, uuid=earlier.uuid, selected_branch=None)
        # only pass the conflict selection to the combined conflict in this one case
        if (
            earlier.diff_branch_value == later.diff_branch_value
            and earlier.base_branch_value == later.base_branch_value
            and earlier.selected_branch is not None
            and later.selected_branch is None
        ):
            combined.selected_branch = earlier.selected_branch
        else:
            combined.selected_branch = later.selected_branch
        return combined

    def _combine_properties(
        self, earlier_properties: set[EnrichedDiffProperty], later_properties: set[EnrichedDiffProperty]
    ) -> set[EnrichedDiffProperty]:
        earlier_props_by_type = {prop.property_type: prop for prop in earlier_properties}
        later_props_by_type = {prop.property_type: prop for prop in later_properties}
        common_property_types = set(earlier_props_by_type.keys()) & set(later_props_by_type.keys())
        combined_properties: set[EnrichedDiffProperty] = set()
        for earlier_property in earlier_properties:
            if earlier_property.property_type not in common_property_types:
                combined_properties.add(deepcopy(earlier_property))
                continue
            later_property = later_props_by_type[earlier_property.property_type]
            if earlier_property.action is DiffAction.ADDED and later_property.action is DiffAction.REMOVED:
                continue
            combined_action = self._combine_actions(earlier=earlier_property.action, later=later_property.action)
            if earlier_property.previous_value == later_property.new_value or {
                earlier_property.previous_value,
                later_property.new_value,
            } <= {None, NULL_VALUE}:
                combined_action = DiffAction.UNCHANGED
            combined_conflict = self.combine_conflicts(earlier=earlier_property.conflict, later=later_property.conflict)
            combined_properties.add(
                replace(
                    later_property,
                    previous_label=earlier_property.previous_label,
                    previous_value=earlier_property.previous_value,
                    action=combined_action,
                    conflict=combined_conflict,
                )
            )
        combined_properties |= {
            deepcopy(prop) for prop in later_properties if prop.property_type not in common_property_types
        }
        return combined_properties

    def _combine_attributes(
        self, earlier_attributes: set[EnrichedDiffAttribute], later_attributes: set[EnrichedDiffAttribute]
    ) -> set[EnrichedDiffAttribute]:
        earlier_attrs_by_name = {attr.name: attr for attr in earlier_attributes}
        later_attrs_by_name = {attr.name: attr for attr in later_attributes}
        common_attr_names = set(earlier_attrs_by_name.keys()) & set(later_attrs_by_name.keys())
        combined_attributes: set[EnrichedDiffAttribute] = set()
        for earlier_attribute in earlier_attributes:
            if earlier_attribute.name not in common_attr_names:
                combined_attributes.add(deepcopy(earlier_attribute))
                continue
            later_attribute = later_attrs_by_name[earlier_attribute.name]
            if not self._should_include(earlier=earlier_attribute.action, later=later_attribute.action):
                continue
            combined_properties = self._combine_properties(
                earlier_properties=earlier_attribute.properties, later_properties=later_attribute.properties
            )
            if all(p.action is DiffAction.UNCHANGED for p in combined_properties):
                combined_action = DiffAction.UNCHANGED
            else:
                combined_action = self._combine_actions(earlier=earlier_attribute.action, later=later_attribute.action)
            if combined_properties:
                combined_attribute = EnrichedDiffAttribute(
                    name=later_attribute.name,
                    changed_at=later_attribute.changed_at,
                    action=combined_action,
                    path_identifier=later_attribute.path_identifier,
                    properties=combined_properties,
                )
                combined_attributes.add(combined_attribute)
        combined_attributes |= {
            deepcopy(attribute) for attribute in later_attributes if attribute.name not in common_attr_names
        }
        return combined_attributes

    def _combine_cardinality_one_relationship_elements(
        self, elements: Iterable[EnrichedDiffSingleRelationship]
    ) -> EnrichedDiffSingleRelationship:
        ordered_elements = sorted(elements, key=lambda e: e.changed_at)
        if len(ordered_elements) < 2:
            return deepcopy(next(iter(elements)))
        combined_action = ordered_elements[0].action
        combined_properties = ordered_elements[0].properties
        for element in ordered_elements[1:]:
            combined_action = self._combine_actions(earlier=combined_action, later=element.action)
            combined_properties = self._combine_properties(
                earlier_properties=combined_properties, later_properties=element.properties
            )
        if all(p.action is DiffAction.UNCHANGED for p in combined_properties):
            combined_action = DiffAction.UNCHANGED
        final_element = ordered_elements[-1]
        peer_id = final_element.peer_id
        peer_label = final_element.peer_label
        # if this relationship is removed and was updated earlier, use the previous peer ID from the update
        if combined_action is DiffAction.REMOVED:
            for element in ordered_elements:
                for prop in element.properties:
                    if (
                        prop.property_type is DatabaseEdgeType.IS_RELATED
                        and prop.action is DiffAction.UPDATED
                        and prop.previous_value
                    ):
                        peer_id = prop.previous_value
                        peer_label = prop.previous_label
                        break
        return EnrichedDiffSingleRelationship(
            changed_at=final_element.changed_at,
            action=combined_action,
            peer_id=peer_id,
            peer_label=peer_label,
            path_identifier=final_element.path_identifier,
            properties=combined_properties,
            conflict=self.combine_conflicts(earlier=ordered_elements[0].conflict, later=final_element.conflict),
        )

    def _combined_cardinality_many_relationship_elements(
        self, earlier_elements: set[EnrichedDiffSingleRelationship], later_elements: set[EnrichedDiffSingleRelationship]
    ) -> set[EnrichedDiffSingleRelationship]:
        earlier_elements_by_peer_id = {element.peer_id: element for element in earlier_elements}
        later_elements_by_peer_id = {element.peer_id: element for element in later_elements}
        common_peer_ids = set(earlier_elements_by_peer_id.keys()) & set(later_elements_by_peer_id.keys())
        combined_elements: set[EnrichedDiffSingleRelationship] = set()
        for earlier_element in earlier_elements:
            if earlier_element.peer_id not in common_peer_ids:
                combined_elements.add(deepcopy(earlier_element))
                continue
            later_element = later_elements_by_peer_id[earlier_element.peer_id]
            if not self._should_include(earlier=earlier_element.action, later=later_element.action):
                continue
            combined_properties = self._combine_properties(
                earlier_properties=earlier_element.properties, later_properties=later_element.properties
            )
            if all(p.action is DiffAction.UNCHANGED for p in combined_properties):
                combined_action = DiffAction.UNCHANGED
            else:
                combined_action = self._combine_actions(earlier=earlier_element.action, later=later_element.action)
            combined_element = EnrichedDiffSingleRelationship(
                changed_at=later_element.changed_at,
                action=combined_action,
                peer_id=later_element.peer_id,
                peer_label=later_element.peer_label,
                path_identifier=later_element.path_identifier,
                properties=combined_properties,
            )
            combined_elements.add(combined_element)
        combined_elements |= {
            deepcopy(later_element) for later_element in later_elements if later_element.peer_id not in common_peer_ids
        }
        return combined_elements

    def _combine_relationships(
        self,
        earlier_relationships: set[EnrichedDiffRelationship],
        later_relationships: set[EnrichedDiffRelationship],
        node_id: str,
    ) -> set[EnrichedDiffRelationship]:
        earlier_rels_by_name = {rel.name: rel for rel in earlier_relationships}
        later_rels_by_name = {rel.name: rel for rel in later_relationships}
        common_rel_names = set(earlier_rels_by_name.keys()) & set(later_rels_by_name.keys())
        combined_relationships: set[EnrichedDiffRelationship] = set()
        for earlier_relationship in earlier_relationships:
            if earlier_relationship.name not in common_rel_names:
                copied = deepcopy(earlier_relationship)
                copied.nodes = set()
                combined_relationships.add(copied)
                continue
            later_relationship = later_rels_by_name[earlier_relationship.name]
            if len(earlier_relationship.relationships) == 0 and len(later_relationship.relationships) == 0:
                combined_relationship_elements = set()
            elif earlier_relationship.cardinality is RelationshipCardinality.ONE:
                combined_relationship_elements = {
                    self._combine_cardinality_one_relationship_elements(
                        elements=(earlier_relationship.relationships | later_relationship.relationships)
                    )
                }
            else:
                combined_relationship_elements = self._combined_cardinality_many_relationship_elements(
                    earlier_elements=earlier_relationship.relationships, later_elements=later_relationship.relationships
                )
            combined_relationship_elements = {cre for cre in combined_relationship_elements if cre.properties}
            if all(cre.action is DiffAction.UNCHANGED for cre in combined_relationship_elements):
                combined_action = DiffAction.UNCHANGED
            else:
                combined_action = self._combine_actions(
                    earlier=earlier_relationship.action, later=later_relationship.action
                )
            parent_rel_name = self._get_parent_relationship_name(node_id=node_id)
            includes_parent = parent_rel_name == later_relationship.name
            if combined_relationship_elements or includes_parent:
                combined_relationship = EnrichedDiffRelationship(
                    name=later_relationship.name,
                    label=later_relationship.label,
                    identifier=later_relationship.identifier,
                    cardinality=later_relationship.cardinality,
                    changed_at=later_relationship.changed_at or earlier_relationship.changed_at,
                    action=combined_action,
                    path_identifier=later_relationship.path_identifier,
                    relationships=combined_relationship_elements,
                    nodes=set(),
                )
                combined_relationships.add(combined_relationship)
        for later_relationship in later_relationships:
            if later_relationship.name in common_rel_names:
                continue
            copied = deepcopy(later_relationship)
            copied.nodes = set()
            combined_relationships.add(copied)
        return combined_relationships

    def _copy_node_without_parents(self, node: EnrichedDiffNode) -> EnrichedDiffNode:
        rels_without_parents = {replace(r, nodes=set()) for r in node.relationships}
        for rel in rels_without_parents:
            rel.reset_summaries()
        node_without_parents = replace(node, relationships=rels_without_parents)
        return deepcopy(node_without_parents)

    def _combine_nodes(self, node_pairs: list[NodePair]) -> set[EnrichedDiffNode]:
        combined_nodes: set[EnrichedDiffNode] = set()
        for node_pair in node_pairs:
            if node_pair.earlier is None:
                if node_pair.later is not None:
                    copied = self._copy_node_without_parents(node_pair.later)
                    combined_nodes.add(copied)
                continue
            if node_pair.later is None:
                if node_pair.earlier is not None:
                    copied = self._copy_node_without_parents(node_pair.earlier)
                    combined_nodes.add(copied)
                continue
            combined_attributes = self._combine_attributes(
                earlier_attributes=node_pair.earlier.attributes, later_attributes=node_pair.later.attributes
            )
            combined_relationships = self._combine_relationships(
                earlier_relationships=node_pair.earlier.relationships,
                later_relationships=node_pair.later.relationships,
                node_id=node_pair.later.uuid,
            )
            if all(ca.action is DiffAction.UNCHANGED for ca in combined_attributes) and all(
                cr.action is DiffAction.UNCHANGED for cr in combined_relationships
            ):
                combined_action = DiffAction.UNCHANGED
            else:
                combined_action = self._combine_actions(earlier=node_pair.earlier.action, later=node_pair.later.action)
            combined_conflict = self.combine_conflicts(
                earlier=node_pair.earlier.conflict, later=node_pair.later.conflict
            )
            if (
                combined_attributes
                or combined_relationships
                or combined_conflict
                or node_pair.later.uuid in self._parent_node_uuids
            ):
                combined_nodes.add(
                    EnrichedDiffNode(
                        uuid=node_pair.later.uuid,
                        kind=node_pair.later.kind,
                        label=node_pair.later.label,
                        changed_at=node_pair.later.changed_at or node_pair.earlier.changed_at,
                        action=combined_action,
                        path_identifier=node_pair.later.path_identifier,
                        attributes=combined_attributes,
                        relationships=combined_relationships,
                        conflict=combined_conflict,
                    )
                )
        return combined_nodes

    def _link_child_nodes(self, nodes: Iterable[EnrichedDiffNode]) -> None:
        nodes_by_uuid: dict[str, EnrichedDiffNode] = {n.uuid: n for n in nodes}
        for child_node in nodes_by_uuid.values():
            if child_node.uuid not in self._child_parent_uuid_map:
                continue
            parent_uuid, parent_rel_name = self._child_parent_uuid_map[child_node.uuid]
            parent_node = nodes_by_uuid[parent_uuid]
            parent_rel = child_node.get_relationship(name=parent_rel_name)
            parent_rel.nodes.add(parent_node)

    async def combine(self, earlier_diffs: EnrichedDiffs, later_diffs: EnrichedDiffs) -> EnrichedDiffs:
        combined_diffs: list[EnrichedDiffRoot] = []
        for earlier, later in (
            (earlier_diffs.base_branch_diff, later_diffs.base_branch_diff),
            (earlier_diffs.diff_branch_diff, later_diffs.diff_branch_diff),
        ):
            self._initialize(earlier_diff=earlier, later_diff=later)
            filtered_node_pairs = self._filter_nodes_to_keep(earlier_diff=earlier, later_diff=later)
            combined_nodes = self._combine_nodes(node_pairs=filtered_node_pairs)
            self._link_child_nodes(nodes=combined_nodes)
            if earlier.exists_on_database:
                diff_uuid = earlier.uuid
                partner_uuid = earlier.partner_uuid
            else:
                diff_uuid = later.uuid
                partner_uuid = later.partner_uuid
            combined_diffs.append(
                EnrichedDiffRoot(
                    uuid=diff_uuid,
                    partner_uuid=partner_uuid,
                    base_branch_name=later.base_branch_name,
                    diff_branch_name=later.diff_branch_name,
                    from_time=earlier.from_time,
                    to_time=later.to_time,
                    tracking_id=later.tracking_id,
                    nodes=combined_nodes,
                )
            )
        base_branch_diff, diff_branch_diff = combined_diffs
        base_branch_diff.partner_uuid = diff_branch_diff.uuid
        diff_branch_diff.partner_uuid = base_branch_diff.uuid
        return EnrichedDiffs(
            base_branch_name=later_diffs.base_branch_name,
            diff_branch_name=later_diffs.diff_branch_name,
            base_branch_diff=base_branch_diff,
            diff_branch_diff=diff_branch_diff,
        )



from pydantic import BaseModel, Field

from infrahub.context import InfrahubContext


class CheckArtifactCreate(BaseModel):
    """Runs a check to verify the creation of an artifact."""

    artifact_name: str = Field(..., description="Name of the artifact")
    artifact_definition: str = Field(..., description="The the ID of the artifact definition")
    commit: str = Field(..., description="The commit to target")
    content_type: str = Field(..., description="Content type of the artifact")
    transform_type: str = Field(..., description="The type of transform associated with this artifact")
    transform_location: str = Field(..., description="The transforms location within the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this artifact")
    target_kind: str = Field(..., description="The kind of the target object for this artifact")
    target_name: str = Field(..., description="Name of the artifact target")
    artifact_id: str | None = Field(default=None, description="The id of the artifact if it previously existed")
    query: str = Field(..., description="The name of the query to use when collecting data")
    timeout: int = Field(..., description="Timeout for requests used to generate this artifact")
    variables: dict = Field(..., description="Input variables when generating the artifact")
    validator_id: str = Field(..., description="The ID of the validator")
    context: InfrahubContext = Field(..., description="The context of the task")



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality

from ...attribute_schema import AttributeSchema as Attr
from ...dropdown import DropdownChoice
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

builtin_ipam = GenericSchema(
    name="IPNamespace",
    namespace="Builtin",
    label="IP Namespace",
    description="A generic container for IP prefixes and IP addresses",
    include_in_menu=False,
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    icon="mdi:format-list-group",
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    generate_profile=False,
    attributes=[
        Attr(name="name", kind="Text", unique=True, branch=BranchSupportType.AWARE, order_weight=1000),
        Attr(name="description", kind="Text", optional=True, branch=BranchSupportType.AWARE, order_weight=2000),
    ],
    relationships=[
        Rel(
            name="ip_prefixes",
            label="IP Prefixes",
            peer=InfrahubKind.IPPREFIX,
            identifier="ip_namespace__ip_prefix",
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
            allow_override=AllowOverrideType.NONE,
        ),
        Rel(
            name="ip_addresses",
            label="IP Addresses",
            peer=InfrahubKind.IPADDRESS,
            identifier="ip_namespace__ip_address",
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
            allow_override=AllowOverrideType.NONE,
        ),
    ],
)

builtin_ip_prefix = GenericSchema(
    name="IPPrefix",
    label="IP Prefix",
    namespace="Builtin",
    description="IPv6 or IPv4 prefix also referred as network",
    include_in_menu=False,
    default_filter="prefix__value",
    order_by=["prefix__version", "prefix__binary_address", "prefix__prefixlen"],
    display_labels=["prefix__value"],
    icon="mdi:ip-network",
    branch=BranchSupportType.AWARE,
    hierarchical=True,
    attributes=[
        Attr(name="prefix", kind="IPNetwork", branch=BranchSupportType.AWARE, order_weight=1000),
        Attr(name="description", kind="Text", optional=True, branch=BranchSupportType.AWARE, order_weight=2000),
        Attr(
            name="member_type",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name="prefix",
                    label="Prefix",
                    description="Prefix serves as container for other prefixes",
                ),
                DropdownChoice(
                    name="address",
                    label="Address",
                    description="Prefix serves as subnet for IP addresses",
                ),
            ],
            branch=BranchSupportType.AWARE,
            default_value="address",
            order_weight=3000,
        ),
        Attr(
            name="is_pool",
            kind="Boolean",
            branch=BranchSupportType.AWARE,
            default_value=False,
            order_weight=4000,
            description="All IP addresses within this prefix are considered usable",
        ),
        Attr(name="is_top_level", kind="Boolean", read_only=True, optional=True, allow_override=AllowOverrideType.NONE),
        Attr(name="utilization", kind="Number", read_only=True, optional=True, allow_override=AllowOverrideType.NONE),
        Attr(name="netmask", kind="Text", read_only=True, optional=True, allow_override=AllowOverrideType.NONE),
        Attr(name="hostmask", kind="Text", read_only=True, optional=True, allow_override=AllowOverrideType.NONE),
        Attr(name="network_address", kind="Text", read_only=True, optional=True, allow_override=AllowOverrideType.NONE),
        Attr(
            name="broadcast_address", kind="Text", read_only=True, optional=True, allow_override=AllowOverrideType.NONE
        ),
    ],
    relationships=[
        Rel(
            name="ip_namespace",
            label="IP Namespace",
            peer=InfrahubKind.IPNAMESPACE,
            identifier="ip_namespace__ip_prefix",
            optional=True,
            cardinality=Cardinality.ONE,
            allow_override=AllowOverrideType.NONE,
        ),
        Rel(
            name="ip_addresses",
            label="IP Addresses",
            peer=InfrahubKind.IPADDRESS,
            identifier="ip_prefix__ip_address",
            optional=True,
            cardinality=Cardinality.MANY,
            allow_override=AllowOverrideType.NONE,
            read_only=True,
        ),
        Rel(
            name="resource_pool",
            peer="CoreIPAddressPool",
            identifier="ipaddresspool__resource",
            cardinality=Cardinality.MANY,
            branch=BranchSupportType.AGNOSTIC,
            optional=True,
            read_only=True,
        ),
    ],
)

builtin_ip_address = GenericSchema(
    name="IPAddress",
    label="IP Address",
    namespace="Builtin",
    description="IPv6 or IPv4 address",
    include_in_menu=False,
    default_filter="address__value",
    order_by=["address__version", "address__binary_address"],
    display_labels=["address__value"],
    icon="mdi:ip-outline",
    branch=BranchSupportType.AWARE,
    attributes=[
        Attr(name="address", kind="IPHost", branch=BranchSupportType.AWARE, order_weight=1000),
        Attr(name="description", kind="Text", optional=True, branch=BranchSupportType.AWARE, order_weight=2000),
    ],
    relationships=[
        Rel(
            name="ip_namespace",
            label="IP Namespace",
            peer=InfrahubKind.IPNAMESPACE,
            identifier="ip_namespace__ip_address",
            optional=True,
            cardinality=Cardinality.ONE,
            allow_override=AllowOverrideType.NONE,
        ),
        Rel(
            name="ip_prefix",
            label="IP Prefix",
            peer=InfrahubKind.IPPREFIX,
            identifier="ip_prefix__ip_address",
            optional=True,
            cardinality=Cardinality.ONE,
            allow_override=AllowOverrideType.NONE,
            read_only=True,
        ),
    ],
)

core_ipam_namespace = NodeSchema(
    name="Namespace",
    namespace="Ipam",
    description="A namespace that segments IPAM",
    label="IPAM Namespace",
    default_filter="name__value",
    human_friendly_id=["name__value"],
    order_by=["name__value"],
    display_labels=["name__value"],
    include_in_menu=False,
    icon="mdi:format-list-group",
    branch=BranchSupportType.AWARE,
    inherit_from=[InfrahubKind.IPNAMESPACE],
    attributes=[Attr(name="default", kind="Boolean", optional=True, read_only=True, order_weight=9000)],
)



from infrahub.core.constants import TaskConclusion

LOG_LEVEL_MAPPING = {10: "debug", 20: "info", 30: "warning", 40: "error", 50: "critical"}

CONCLUSION_STATE_MAPPING: dict[str, TaskConclusion] = {
    "Scheduled": TaskConclusion.UNKNOWN,
    "Pending": TaskConclusion.UNKNOWN,
    "Running": TaskConclusion.UNKNOWN,
    "Completed": TaskConclusion.SUCCESS,
    "Failed": TaskConclusion.FAILURE,
    "Cancelled": TaskConclusion.FAILURE,
    "Crashed": TaskConclusion.FAILURE,
    "Paused": TaskConclusion.UNKNOWN,
    "Cancelling": TaskConclusion.FAILURE,
}



from __future__ import annotations

import asyncio
import functools
import importlib
import logging
import platform
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional

import typer
import ujson
from rich.console import Console
from rich.layout import Layout
from rich.logging import RichHandler
from rich.panel import Panel
from rich.pretty import Pretty
from rich.table import Table

from .. import __version__ as sdk_version
from ..async_typer import AsyncTyper
from ..code_generator import CodeGenerator
from ..ctl import config
from ..ctl.branch import app as branch_app
from ..ctl.check import run as run_check
from ..ctl.client import initialize_client, initialize_client_sync
from ..ctl.exceptions import QueryNotFoundError
from ..ctl.generator import run as run_generator
from ..ctl.menu import app as menu_app
from ..ctl.object import app as object_app
from ..ctl.render import list_jinja2_transforms, print_template_errors
from ..ctl.repository import app as repository_app
from ..ctl.repository import get_repository_config
from ..ctl.schema import app as schema_app
from ..ctl.transform import list_transforms
from ..ctl.utils import (
    catch_exception,
    execute_graphql_query,
    load_yamlfile_from_disk_and_exit,
    parse_cli_vars,
)
from ..ctl.validate import app as validate_app
from ..exceptions import GraphQLError, ModuleImportError
from ..schema import MainSchemaTypesAll, SchemaRoot
from ..template import Jinja2Template
from ..template.exceptions import JinjaTemplateError
from ..utils import get_branch, write_to_file
from ..yaml import SchemaFile
from .exporter import dump
from .importer import load
from .parameters import CONFIG_PARAM

if TYPE_CHECKING:
    from ..schema.repository import InfrahubRepositoryConfig

app = AsyncTyper(pretty_exceptions_show_locals=False)

app.add_typer(branch_app, name="branch")
app.add_typer(schema_app, name="schema")
app.add_typer(validate_app, name="validate")
app.add_typer(repository_app, name="repository")
app.add_typer(menu_app, name="menu")
app.add_typer(object_app, name="object", hidden=True)

app.command(name="dump")(dump)
app.command(name="load")(load)

console = Console()


@app.command(name="check")
@catch_exception(console=console)
def check(
    check_name: str = typer.Argument(default="", help="Name of the Python check"),
    branch: Optional[str] = None,
    path: str = typer.Option(".", help="Root directory"),
    debug: bool = False,
    format_json: bool = False,
    _: str = CONFIG_PARAM,
    list_available: bool = typer.Option(False, "--list", help="Show available Python checks"),
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
) -> None:
    """Execute user-defined checks."""

    variables_dict = parse_cli_vars(variables)
    run_check(
        path=path,
        debug=debug,
        branch=branch,
        format_json=format_json,
        list_available=list_available,
        name=check_name,
        variables=variables_dict,
    )


@app.command(name="generator")
@catch_exception(console=console)
async def generator(
    generator_name: str = typer.Argument(default="", help="Name of the Generator"),
    branch: Optional[str] = None,
    path: str = typer.Option(".", help="Root directory"),
    debug: bool = False,
    _: str = CONFIG_PARAM,
    list_available: bool = typer.Option(False, "--list", help="Show available Generators"),
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
) -> None:
    """Run a generator script."""
    await run_generator(
        generator_name=generator_name,
        branch=branch,
        path=path,
        debug=debug,
        list_available=list_available,
        variables=variables,
    )


@app.command(name="run")
@catch_exception(console=console)
async def run(
    script: Path,
    method: str = "run",
    debug: bool = False,
    _: str = CONFIG_PARAM,
    branch: str = typer.Option(None, help="Branch on which to run the script."),
    concurrent: Optional[int] = typer.Option(
        None,
        help="Maximum number of requests to execute at the same time.",
        envvar="INFRAHUB_MAX_CONCURRENT_EXECUTION",
    ),
    timeout: int = typer.Option(60, help="Timeout in sec", envvar="INFRAHUB_TIMEOUT"),
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
) -> None:
    """Execute a script."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)

    log_level = "DEBUG" if debug else "INFO"
    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    log = logging.getLogger("infrahubctl")

    variables_dict = parse_cli_vars(variables)

    directory_name = str(script.parent)
    module_name = script.stem

    if directory_name not in sys.path:
        sys.path.append(directory_name)

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        raise typer.Abort(f"Unable to Load the Python script at {script}") from exc

    if not hasattr(module, method):
        raise typer.Abort(f"Unable to Load the method {method} in the Python script at {script}")

    client = initialize_client(
        branch=branch,
        timeout=timeout,
        max_concurrent_execution=concurrent,
        identifier=module_name,
    )
    func = getattr(module, method)
    await func(client=client, log=log, branch=branch, **variables_dict)


async def render_jinja2_template(template_path: Path, variables: dict[str, Any], data: dict[str, Any]) -> str:
    variables["data"] = data
    jinja_template = Jinja2Template(template=Path(template_path), template_directory=Path())
    try:
        rendered_tpl = await jinja_template.render(variables=variables)
    except JinjaTemplateError as exc:
        print_template_errors(error=exc, console=console)
        raise typer.Exit(1) from exc

    return rendered_tpl


async def _run_transform(
    query_name: str,
    variables: dict[str, Any],
    transform_func: Callable,
    branch: str,
    debug: bool,
    repository_config: InfrahubRepositoryConfig,
) -> Any:
    """
    Query GraphQL for the required data then run a transform on that data.

    Args:
        query_name: Name of the query to load (e.g. tags_query)
        variables: Dictionary of variables used for graphql query
        transform_func: The function responsible for transforming data received from graphql
        branch: Name of the *infrahub* branch that should be queried for data
        debug: Prints debug info to the command line
        repository_config: Repository config object. This is used to load the graphql query from the repository.
    """
    branch = get_branch(branch)

    try:
        response = execute_graphql_query(
            query=query_name,
            variables_dict=variables,
            branch=branch,
            debug=debug,
            repository_config=repository_config,
        )

        # TODO: response is a dict and can't be printed to the console in this way.
        # if debug:
        #     message = ("-" * 40, f"Response for GraphQL Query {query_name}", response, "-" * 40)
        #     console.print("\n".join(message))
    except QueryNotFoundError as exc:
        console.print(f"[red]Unable to find query : {exc}")
        raise typer.Exit(1) from exc
    except GraphQLError as exc:
        console.print(f"[red]{len(exc.errors)} error(s) occurred while executing the query")
        for error in exc.errors:
            if isinstance(error, dict) and "message" in error and "locations" in error:
                console.print(f"[yellow] - Message: {error['message']}")  # type: ignore[typeddict-item]
                console.print(f"[yellow]   Location: {error['locations']}")  # type: ignore[typeddict-item]
            elif isinstance(error, str) and "Branch:" in error:
                console.print(f"[yellow] - {error}")
                console.print("[yellow]   you can specify a different branch with --branch")
        raise typer.Abort()

    if asyncio.iscoroutinefunction(transform_func):
        output = await transform_func(response)
    else:
        output = transform_func(response)
    return output


@app.command(name="render")
@catch_exception(console=console)
async def render(
    transform_name: str = typer.Argument(default="", help="Name of the Python transformation", show_default=False),
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
    branch: str = typer.Option(None, help="Branch on which to render the transform."),
    debug: bool = False,
    _: str = CONFIG_PARAM,
    list_available: bool = typer.Option(False, "--list", help="Show available transforms"),
    out: str = typer.Option(None, help="Path to a file to save the result."),
) -> None:
    """Render a local Jinja2 Transform for debugging purpose."""

    variables_dict = parse_cli_vars(variables)
    repository_config = get_repository_config(Path(config.INFRAHUB_REPO_CONFIG_FILE))

    if list_available or not transform_name:
        list_jinja2_transforms(config=repository_config)
        return

    # Load transform config
    try:
        transform_config = repository_config.get_jinja2_transform(name=transform_name)
    except KeyError as exc:
        console.print(f'[red]Unable to find "{transform_name}" in {config.INFRAHUB_REPO_CONFIG_FILE}')
        list_jinja2_transforms(config=repository_config)
        raise typer.Exit(1) from exc

    # Construct transform function used to transform data returned from the API
    transform_func = functools.partial(render_jinja2_template, transform_config.template_path, variables_dict)

    # Query GQL and run the transform
    result = await _run_transform(
        query_name=transform_config.query,
        variables=variables_dict,
        transform_func=transform_func,
        branch=branch,
        debug=debug,
        repository_config=repository_config,
    )

    # Output data
    if out:
        write_to_file(Path(out), result)
    else:
        console.print(result)


@app.command(name="transform")
@catch_exception(console=console)
def transform(
    transform_name: str = typer.Argument(default="", help="Name of the Python transformation", show_default=False),
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
    branch: str = typer.Option(None, help="Branch on which to run the transformation"),
    debug: bool = False,  # noqa: ARG001
    _: str = CONFIG_PARAM,
    list_available: bool = typer.Option(False, "--list", help="Show available transforms"),
    out: str = typer.Option(None, help="Path to a file to save the result."),
) -> None:
    """Render a local transform (TransformPython) for debugging purpose."""

    variables_dict = parse_cli_vars(variables)
    repository_config = get_repository_config(Path(config.INFRAHUB_REPO_CONFIG_FILE))

    if list_available or not transform_name:
        list_transforms(config=repository_config)
        return

    transform_config = repository_config.get_python_transform(name=transform_name)

    # Get client
    client = initialize_client(branch=branch)

    # Get python transform class instance

    relative_path = str(transform_config.file_path.parent) if transform_config.file_path.parent != Path() else None

    try:
        transform_class = transform_config.load_class(import_root=str(Path.cwd()), relative_path=relative_path)
    except ModuleImportError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    transform = transform_class(client=client, branch=branch)
    # Get data
    query_str = repository_config.get_query(name=transform.query).load_query()
    data = asyncio.run(
        transform.client.execute_graphql(query=query_str, variables=variables_dict, branch_name=transform.branch_name)
    )

    # Run Transform
    result = asyncio.run(transform.run(data=data))

    if isinstance(result, str):
        json_string = result
    else:
        json_string = ujson.dumps(result, indent=2, sort_keys=True)

    if out:
        write_to_file(Path(out), json_string)
    else:
        console.print(json_string)


@app.command(name="protocols")
@catch_exception(console=console)
def protocols(
    schemas: list[Path] = typer.Option(None, help="List of schemas or directory to load."),
    branch: str = typer.Option(None, help="Branch of schema to export Python protocols for."),
    sync: bool = typer.Option(False, help="Generate for sync or async."),
    _: str = CONFIG_PARAM,
    out: str = typer.Option("schema_protocols.py", help="Path to a file to save the result."),
) -> None:
    """Export Python protocols corresponding to a schema."""

    schema: dict[str, MainSchemaTypesAll] = {}

    if schemas:
        schemas_data = load_yamlfile_from_disk_and_exit(paths=schemas, file_type=SchemaFile, console=console)

        for data in schemas_data:
            data.load_content()
            schema_root_data = data.content or {}
            schema_root = SchemaRoot(**schema_root_data)
            schema.update({item.kind: item for item in schema_root.nodes + schema_root.generics})

    else:
        client = initialize_client_sync()
        branch = branch or client.default_branch
        schema.update(client.schema.fetch(branch=branch))

    code_generator = CodeGenerator(schema=schema)

    if out:
        output_file = Path(out)
        write_to_file(output_file, code_generator.render(sync=sync))
        console.print(f"Python protocols exported in {output_file}")
    else:
        console.print(code_generator.render(sync=sync))


@app.command(name="version")
@catch_exception(console=console)
def version() -> None:
    """Display the version of Python and the version of the Python SDK in use."""

    console.print(f"Python: {platform.python_version()}\nPython SDK: v{sdk_version}")


@app.command(name="info")
@catch_exception(console=console)
def info(  # noqa: PLR0915
    detail: bool = typer.Option(False, help="Display detailed information."),
    _: str = CONFIG_PARAM,
) -> None:
    """Display the status of the Python SDK."""

    info: dict[str, Any] = {
        "error": None,
        "status": ":x:",
        "infrahub_version": "N/A",
        "user_info": {},
        "groups": {},
    }
    try:
        client = initialize_client_sync()
        info["infrahub_version"] = client.get_version()
        info["user_info"] = client.get_user()
        info["status"] = ":white_heavy_check_mark:"
        info["groups"] = client.get_user_permissions()
    except Exception as e:
        info["error"] = f"{e!s} ({e.__class__.__name__})"

    if detail:
        layout = Layout()

        # Layout structure
        new_console = Console(height=45)
        layout = Layout()
        layout.split_column(
            Layout(name="body", ratio=1),
        )
        layout["body"].split_row(
            Layout(name="left"),
            Layout(name="right"),
        )

        layout["left"].split_column(
            Layout(name="connection_status", size=7),
            Layout(name="client_info", ratio=1),
        )

        layout["right"].split_column(
            Layout(name="version_info", size=7),
            Layout(name="infrahub_info", ratio=1),
        )

        # Connection status panel
        connection_status = Table(show_header=False, box=None)
        connection_status.add_row("Server Address:", client.config.address)
        connection_status.add_row("Status:", info["status"])
        if info["error"]:
            connection_status.add_row("Error Reason:", info["error"])
        layout["connection_status"].update(Panel(connection_status, title="Connection Status"))

        # Version information panel
        version_info = Table(show_header=False, box=None)
        version_info.add_row("Python Version:", platform.python_version())
        version_info.add_row("Infrahub Version", info["infrahub_version"])
        version_info.add_row("Infrahub SDK:", sdk_version)
        layout["version_info"].update(Panel(version_info, title="Version Information"))

        # SDK client configuration panel
        pretty_model = Pretty(client.config.model_dump(), expand_all=True)
        layout["client_info"].update(Panel(pretty_model, title="Client Info"))

        # Infrahub information planel
        infrahub_info = Table(show_header=False, box=None)
        if info["user_info"]:
            infrahub_info.add_row("User:", info["user_info"]["AccountProfile"]["display_label"])
            infrahub_info.add_row(
                "Description:",
                info["user_info"]["AccountProfile"]["description"]["value"],
            )
            infrahub_info.add_row("Status:", info["user_info"]["AccountProfile"]["status"]["label"])
            infrahub_info.add_row(
                "Number of Groups:",
                str(info["user_info"]["AccountProfile"]["member_of_groups"]["count"]),
            )

            if groups := info["groups"]:
                infrahub_info.add_row("Groups:", "")
                for group, roles in groups.items():
                    infrahub_info.add_row("", group, ", ".join(roles))

        layout["infrahub_info"].update(Panel(infrahub_info, title="Infrahub Info"))

        new_console.print(layout)
    else:
        # Simple output
        table = Table(show_header=False, box=None)
        table.add_row("Address:", client.config.address)
        table.add_row("Connection Status:", info["status"])
        if info["error"]:
            table.add_row("Connection Error:", info["error"])

        table.add_row("Python Version:", platform.python_version())
        table.add_row("SDK Version:", sdk_version)
        table.add_row("Infrahub Version:", info["infrahub_version"])
        if account := info["user_info"].get("AccountProfile"):
            table.add_row("User:", account["display_label"])

        console.print(table)



from typing import Any, AsyncGenerator

from graphene import Field, Int, ObjectType, Schema, String
from graphene.types.generic import GenericScalar
from graphql import GraphQLResolveInfo

from .graphql_query import resolver_graphql_query

GraphQLQuerySubscription = Field(
    GenericScalar(),
    name=String(),
    params=GenericScalar(required=False),
    interval=Int(required=False),
)


class InfrahubBaseSubscription(ObjectType):
    query = GraphQLQuerySubscription
    graphql_schema: Schema | None = None

    @classmethod
    async def subscribe_query(
        cls,
        parent: dict,
        info: GraphQLResolveInfo,
        name: str,
        params: dict[str, Any] | None = None,
        interval: int | None = 10,
    ) -> AsyncGenerator[dict[str, Any], None]:
        if not cls.graphql_schema:
            raise RuntimeError("Subscription initialized without graphql schema")
        if not interval:
            interval = 10
        async for result in resolver_graphql_query(
            parent=parent, info=info, name=name, graphql_schema=cls.graphql_schema, params=params, interval=interval
        ):
            yield result



import sys

try:
    from .cli_commands import app
except ImportError as exc:
    sys.exit(
        f"Module {exc.name} is not available, install the 'ctl' extra of the infrahub-sdk package, `pip install 'infrahub-sdk[ctl]'` or enable the "
        "Poetry shell and run `poetry install --extras ctl`."
    )

__all__ = ["app"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from ..shared import AttributeSchemaMigration, SchemaMigration


class MigrationQuery(Query):
    type: QueryType = QueryType.WRITE

    def __init__(
        self,
        migration: SchemaMigration,
        **kwargs: Any,
    ):
        self.migration = migration
        super().__init__(**kwargs)

    def get_nbr_migrations_executed(self) -> int:
        return self.num_of_results


class AttributeMigrationQuery(Query):
    type: QueryType = QueryType.WRITE

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        self.migration = migration
        super().__init__(**kwargs)

    def get_nbr_migrations_executed(self) -> int:
        return self.num_of_results



from neo4j.exceptions import Neo4jError
from prefect import task
from prefect.cache_policies import NONE

from infrahub.core import utils
from infrahub.core.graph.schema import GRAPH_SCHEMA
from infrahub.core.query import QueryType
from infrahub.database import DatabaseType, InfrahubDatabase

from .models import TelemetryDatabaseData, TelemetryDatabaseServerData, TelemetryDatabaseSystemInfoData


async def get_server_info(db: InfrahubDatabase) -> list[TelemetryDatabaseServerData]:
    data: list[TelemetryDatabaseServerData] = []

    try:
        results = await db.execute_query(query="SHOW SERVERS YIELD *", name="get_server_info", type=QueryType.READ)
    except Neo4jError:
        return []

    for result in results:
        data.append(
            TelemetryDatabaseServerData(
                name=result["name"],
                version=result["version"],
            )
        )

    return data


async def get_system_info(db: InfrahubDatabase) -> TelemetryDatabaseSystemInfoData:
    query = """
    CALL dbms.queryJmx("java.lang:type=OperatingSystem")
    YIELD attributes
    RETURN
        attributes.AvailableProcessors as processor_available,
        attributes.TotalMemorySize as memory_total,
        attributes.FreeMemorySize as memory_available
    """
    results = await db.execute_query(query=query, name="get_system_info", type=QueryType.READ)

    return TelemetryDatabaseSystemInfoData(
        memory_total=results[0]["memory_total"]["value"],
        memory_available=results[0]["memory_available"]["value"],
        processor_available=results[0]["processor_available"]["value"],
    )


@task(name="telemetry-gather-db", task_run_name="Gather Database Information", cache_policy=NONE)
async def gather_database_information(db: InfrahubDatabase) -> TelemetryDatabaseData:
    async with db.start_session() as dbs:
        server_info = []
        system_info = None
        database_type = db.db_type.value

        if db.db_type == DatabaseType.NEO4J:
            server_info = await get_server_info(db=dbs)
            system_info = await get_system_info(db=dbs)

            # server_info is only available on Neo4j Enterprise
            #  so if it's not empty, we can assume the database is of type Enterprise
            if len(server_info) == 0:
                database_type = f"{database_type}-community"
            else:
                database_type = f"{database_type}-enterprise"

        data = TelemetryDatabaseData(
            database_type=database_type,
            relationship_count={
                "total": await utils.count_relationships(db=dbs),
            },
            node_count={
                "total": await utils.count_nodes(db=dbs),
            },
            servers=server_info,
            system_info=system_info,
        )

        for name in GRAPH_SCHEMA["relationships"]:
            data.relationship_count[name] = await utils.count_relationships(db=dbs, label=name)

        for name in GRAPH_SCHEMA["nodes"]:
            data.node_count[name] = await utils.count_nodes(db=dbs, label=name)

        return data



from typing import TYPE_CHECKING

from infrahub import lock
from infrahub.core import registry
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from infrahub.core.branch import Branch

log = get_logger()


async def refresh_branches(db: InfrahubDatabase) -> None:
    """Pull all the branches from the database and update the registry.

    If a branch is already present with a different value for the hash
    We pull the new schema from the database and we update the registry.
    """
    from infrahub.graphql.manager import GraphQLSchemaManager

    async with lock.registry.local_schema_lock():
        branches = await registry.branch_object.get_list(db=db)
        for new_branch in branches:
            if new_branch.name in registry.branch:
                branch_registry: Branch = registry.branch[new_branch.name]
                if (
                    branch_registry.schema_hash
                    and branch_registry.schema_hash.main != new_branch.active_schema_hash.main
                ):
                    log.info(
                        "New hash detected",
                        branch=new_branch.name,
                        hash_current=branch_registry.schema_hash.main,
                        hash_new=new_branch.active_schema_hash.main,
                        worker=WORKER_IDENTITY,
                    )
                    await registry.schema.load_schema(db=db, branch=new_branch)
                    registry.branch[new_branch.name] = new_branch
                    schema_branch = registry.schema.get_schema_branch(name=new_branch.name)
                    gqlm = GraphQLSchemaManager.get_manager_for_branch(branch=new_branch, schema_branch=schema_branch)
                    gqlm.get_graphql_schema(
                        include_query=True,
                        include_mutation=True,
                        include_subscription=True,
                        include_types=True,
                    )

            else:
                log.info("New branch detected, pulling schema", branch=new_branch.name, worker=WORKER_IDENTITY)
                await registry.schema.load_schema(db=db, branch=new_branch)
                registry.branch[new_branch.name] = new_branch
                schema_branch = registry.schema.get_schema_branch(name=new_branch.name)
                gqlm = GraphQLSchemaManager.get_manager_for_branch(branch=new_branch, schema_branch=schema_branch)
                gqlm.get_graphql_schema(
                    include_query=True,
                    include_mutation=True,
                    include_subscription=True,
                    include_types=True,
                )

        purged_branches = await registry.purge_inactive_branches(db=db, active_branches=branches)
        for branch_name in purged_branches:
            log.info(f"Removed branch {branch_name!r} from the registry", branch=branch_name, worker=WORKER_IDENTITY)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Self

from graphene import Boolean, InputField, InputObjectType, List, Mutation, String
from infrahub_sdk.utils import compare_lists

from infrahub import config
from infrahub.core.account import GlobalPermission, ObjectPermission
from infrahub.core.changelog.models import NodeChangelog, RelationshipChangelogGetter
from infrahub.core.constants import (
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
    RelationshipCardinality,
)
from infrahub.core.manager import NodeManager
from infrahub.core.query.node import NodeGetKindQuery
from infrahub.core.query.relationship import (
    RelationshipGetPeerQuery,
    RelationshipPeerData,
)
from infrahub.core.relationship import Relationship
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.group_action import GroupMemberAddedEvent, GroupMemberRemovedEvent
from infrahub.events.models import EventNode
from infrahub.events.node_action import NodeUpdatedEvent
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.groups.ancestors import collect_ancestors
from infrahub.permissions import get_global_permission_for_kind

from ..types import RelatedNodeInput

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.node import Node
    from infrahub.core.relationship import RelationshipManager
    from infrahub.core.schema.relationship_schema import RelationshipSchema

    from ..initialization import GraphqlContext


RELATIONSHIP_PEERS_TO_IGNORE = [InfrahubKind.NODE]


class GroupUpdateType(str, Enum):
    NONE = "none"
    MEMBERS = "members"
    MEMBER_OF_GROUPS = "member_of_groups"


class RelationshipNodesInput(InputObjectType):
    id = InputField(String(required=True), description="ID of the node at the source of the relationship")
    name = InputField(String(required=True), description="Name of the relationship to add or remove nodes")
    nodes = InputField(
        List(of_type=RelatedNodeInput), description="List of nodes to add or remove to the relationships"
    )


class RelationshipAdd(Mutation):
    class Arguments:
        data = RelationshipNodesInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="relationship_add")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: RelationshipNodesInput,
        context: ContextInput | None = None,
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        relationship_name = str(data.name)

        source = await _validate_node(info=info, data=data)
        nodes = await _validate_peers(info=info, data=data)
        await _validate_permissions(info=info, source_node=source, peers=nodes)
        await _validate_peer_types(info=info, data=data, source_node=source, peers=nodes)

        # This has to be done after validating the permissions
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        rel_schema = source.get_schema().get_relationship(name=relationship_name)
        display_label: str = await source.render_display_label(db=graphql_context.db)
        node_changelog = NodeChangelog(
            node_id=source.get_id(), node_kind=source.get_kind(), display_label=display_label
        )

        existing_peers = await _collect_current_peers(info=info, data=data, source_node=source)

        group_event_type = _get_group_event_type(
            node=source, relationship_schema=rel_schema, relationship_name=relationship_name
        )

        async with graphql_context.db.start_transaction() as db:
            peers: list[EventNode] = []
            for node_data in data.get("nodes"):
                # Instantiate and resolve a relationship
                # This will take care of allocating a node from a pool if needed
                rel = Relationship(schema=rel_schema, branch=graphql_context.branch, node=source)
                await rel.new(db=db, data=node_data)
                await rel.resolve(db=db)
                # Save it only if it does not exist
                if rel.get_peer_id() not in existing_peers.keys():
                    if group_event_type != GroupUpdateType.NONE:
                        peers.append(EventNode(id=rel.get_peer_id(), kind=nodes[rel.get_peer_id()].get_kind()))
                    node_changelog.create_relationship(relationship=rel)
                    await rel.save(db=db)

        if config.SETTINGS.broker.enable and graphql_context.background and node_changelog.has_changes:
            if group_event_type == GroupUpdateType.MEMBERS:
                ancestors = await collect_ancestors(
                    db=graphql_context.db,
                    branch=graphql_context.branch,
                    node_kind=source.get_schema().kind,
                    node_id=source.id,
                )
                group_add_event = GroupMemberAddedEvent(
                    node_id=source.id,
                    kind=source.get_schema().kind,
                    members=peers,
                    ancestors=ancestors,
                    meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                )
                graphql_context.background.add_task(graphql_context.active_service.event.send, group_add_event)

            elif group_event_type == GroupUpdateType.MEMBER_OF_GROUPS:
                group_ids = [node.id for node in peers]
                async with graphql_context.db.start_session() as db:
                    node_kind_query = await NodeGetKindQuery.init(db=db, branch=graphql_context.branch, ids=group_ids)
                    await node_kind_query.execute(db=db)
                    node_kind_map = await node_kind_query.get_node_kind_map()

                    for node_id, node_kind in node_kind_map.items():
                        ancestors = await collect_ancestors(
                            db=graphql_context.db, branch=graphql_context.branch, node_kind=node_kind, node_id=node_id
                        )
                        group_add_event = GroupMemberAddedEvent(
                            node_id=node_id,
                            kind=node_kind,
                            ancestors=ancestors,
                            members=[EventNode(id=source.get_id(), kind=source.get_kind())],
                            meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                        )
                        graphql_context.background.add_task(graphql_context.active_service.event.send, group_add_event)

            else:
                main_event = NodeUpdatedEvent(
                    kind=source.get_schema().kind,
                    node_id=source.id,
                    changelog=node_changelog,
                    fields=[relationship_name],
                    meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                )
                relationship_changelogs = RelationshipChangelogGetter(
                    db=graphql_context.db, branch=graphql_context.branch
                )
                node_changelogs = await relationship_changelogs.get_changelogs(primary_changelog=node_changelog)

                events = [main_event]

                for node_changelog in node_changelogs:
                    meta = EventMeta.from_parent(parent=main_event)
                    event = NodeUpdatedEvent(
                        kind=node_changelog.node_kind,
                        node_id=node_changelog.node_id,
                        changelog=node_changelog,
                        fields=node_changelog.updated_fields,
                        meta=meta,
                    )
                    events.append(event)

                for event in events:
                    graphql_context.background.add_task(graphql_context.active_service.event.send, event)

        return cls(ok=True)


class RelationshipRemove(Mutation):
    class Arguments:
        data = RelationshipNodesInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="relationship_remove")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: RelationshipNodesInput,
        context: ContextInput | None = None,
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        relationship_name = str(data.name)

        source = await _validate_node(info=info, data=data)
        nodes = await _validate_peers(info=info, data=data)
        await _validate_permissions(info=info, source_node=source, peers=nodes)
        await _validate_peer_types(info=info, data=data, source_node=source, peers=nodes)

        # This has to be done after validating the permissions
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        rel_schema = source.get_schema().get_relationship(name=relationship_name)
        display_label: str = await source.render_display_label(db=graphql_context.db)
        node_changelog = NodeChangelog(
            node_id=source.get_id(), node_kind=source.get_kind(), display_label=display_label
        )

        existing_peers = await _collect_current_peers(info=info, data=data, source_node=source)
        group_event_type = _get_group_event_type(
            node=source, relationship_schema=rel_schema, relationship_name=relationship_name
        )

        async with graphql_context.db.start_transaction() as db:
            peers: list[EventNode] = []

            for node_data in data.get("nodes"):
                if node_data.get("id") in existing_peers.keys():
                    # TODO once https://github.com/opsmill/infrahub/issues/792 has been fixed
                    # we should use RelationshipDataDeleteQuery to delete the relationship
                    # it would be more query efficient
                    rel = Relationship(schema=rel_schema, branch=graphql_context.branch, node=source)
                    await rel.load(db=db, data=existing_peers[node_data.get("id")])
                    if group_event_type != GroupUpdateType.NONE:
                        peers.append(EventNode(id=rel.get_peer_id(), kind=nodes[rel.get_peer_id()].get_kind()))
                    node_changelog.delete_relationship(relationship=rel)
                    await rel.delete(db=db)

        if config.SETTINGS.broker.enable and graphql_context.background and node_changelog.has_changes:
            if group_event_type == GroupUpdateType.MEMBERS:
                ancestors = await collect_ancestors(
                    db=graphql_context.db,
                    branch=graphql_context.branch,
                    node_kind=source.get_schema().kind,
                    node_id=source.id,
                )
                group_remove_event = GroupMemberRemovedEvent(
                    node_id=source.id,
                    kind=source.get_schema().kind,
                    members=peers,
                    ancestors=ancestors,
                    meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                )
                graphql_context.background.add_task(graphql_context.active_service.event.send, group_remove_event)
            elif group_event_type == GroupUpdateType.MEMBER_OF_GROUPS:
                group_ids = [node.id for node in peers]
                async with graphql_context.db.start_session() as db:
                    node_kind_query = await NodeGetKindQuery.init(db=db, branch=graphql_context.branch, ids=group_ids)
                    await node_kind_query.execute(db=db)
                    node_kind_map = await node_kind_query.get_node_kind_map()

                    for node_id, node_kind in node_kind_map.items():
                        ancestors = await collect_ancestors(
                            db=graphql_context.db, branch=graphql_context.branch, node_kind=node_kind, node_id=node_id
                        )
                        group_remove_event = GroupMemberRemovedEvent(
                            node_id=node_id,
                            kind=node_kind,
                            members=[EventNode(id=source.get_id(), kind=source.get_kind())],
                            meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                        )
                        graphql_context.background.add_task(
                            graphql_context.active_service.event.send, group_remove_event
                        )
            else:
                main_event = NodeUpdatedEvent(
                    kind=source.get_schema().kind,
                    node_id=source.id,
                    changelog=node_changelog,
                    fields=[relationship_name],
                    meta=EventMeta(branch=graphql_context.branch, context=graphql_context.get_context()),
                )

                relationship_changelogs = RelationshipChangelogGetter(
                    db=graphql_context.db, branch=graphql_context.branch
                )
                node_changelogs = await relationship_changelogs.get_changelogs(primary_changelog=node_changelog)

                events = [main_event]

                for node_changelog in node_changelogs:
                    meta = EventMeta.from_parent(parent=main_event)
                    event = NodeUpdatedEvent(
                        kind=node_changelog.node_kind,
                        node_id=node_changelog.node_id,
                        changelog=node_changelog,
                        fields=node_changelog.updated_fields,
                        meta=meta,
                    )
                    events.append(event)

                for event in events:
                    graphql_context.background.add_task(graphql_context.active_service.event.send, event)

        return cls(ok=True)


async def _validate_node(info: GraphQLResolveInfo, data: RelationshipNodesInput) -> Node:
    graphql_context: GraphqlContext = info.context
    input_id = str(data.id)
    relationship_name = str(data.name)

    if not (
        source := await NodeManager.get_one(
            db=graphql_context.db,
            id=input_id,
            branch=graphql_context.branch,
            include_owner=False,
            include_source=False,
        )
    ):
        raise NodeNotFoundError(node_type="node", identifier=input_id, branch_name=graphql_context.branch.name)

    # Check if the name of the relationship provided exist for this node and is of cardinality Many
    if relationship_name not in source.get_schema().relationship_names:
        raise ValidationError({"name": f"'{relationship_name}' is not a valid relationship for '{source.get_kind()}'"})

    rel_schema = source.get_schema().get_relationship(name=relationship_name)
    if rel_schema.cardinality != RelationshipCardinality.MANY:
        raise ValidationError({"name": f"'{relationship_name}' must be a relationship of cardinality Many"})

    return source


async def _validate_peers(info: GraphQLResolveInfo, data: RelationshipNodesInput) -> dict[str, Node]:
    graphql_context: GraphqlContext = info.context

    # Query the node in the database and validate that all of them exist and are if the correct kind
    node_ids: list[str] = [node_data["id"] for node_data in data.get("nodes") if "id" in node_data]
    nodes = await NodeManager.get_many(
        db=graphql_context.db, ids=node_ids, fields={"display_label": None}, branch=graphql_context.branch
    )
    _, _, in_list2 = compare_lists(list1=list(nodes.keys()), list2=node_ids)
    if in_list2:
        for node_id in in_list2:
            raise ValidationError(f"{node_id!r}: Unable to find the node in the database.")
    return nodes


async def _validate_permissions(info: GraphQLResolveInfo, source_node: Node, peers: dict[str, Node]) -> None:
    graphql_context: GraphqlContext = info.context

    if graphql_context.account_session:
        impacted_schemas = {node.get_schema() for node in [source_node] + list(peers.values())}
        required_permissions: list[GlobalPermission | ObjectPermission] = []
        decision = (
            PermissionDecision.ALLOW_DEFAULT.value
            if graphql_context.branch.is_default
            else PermissionDecision.ALLOW_OTHER.value
        )

        for impacted_schema in impacted_schemas:
            global_action = get_global_permission_for_kind(schema=impacted_schema)

            if global_action:
                required_permissions.append(GlobalPermission(action=global_action, decision=decision))
            else:
                required_permissions.append(
                    ObjectPermission(
                        namespace=impacted_schema.namespace,
                        name=impacted_schema.name,
                        action=PermissionAction.UPDATE.value,
                        decision=decision,
                    )
                )

        graphql_context.active_permissions.raise_for_permissions(permissions=required_permissions)


async def _validate_peer_types(
    info: GraphQLResolveInfo, data: RelationshipNodesInput, source_node: Node, peers: dict[str, Node]
) -> None:
    graphql_context: GraphqlContext = info.context
    relationship_name = str(data.name)
    input_id = str(data.id)
    rel_schema = source_node.get_schema().get_relationship(name=relationship_name)
    for node_id, node in peers.items():
        if rel_schema.peer in RELATIONSHIP_PEERS_TO_IGNORE:
            continue
        if rel_schema.peer not in node.get_labels():
            raise ValidationError(f"{node_id!r} {node.get_kind()!r} is not a valid peer for '{rel_schema.peer}'")

        peer_relationships = [rel for rel in node.get_schema().relationships if rel.identifier == rel_schema.identifier]
        if (
            rel_schema.identifier
            and len(peer_relationships) == 1
            and peer_relationships[0].cardinality == RelationshipCardinality.ONE
        ):
            peer_relationship: RelationshipManager = getattr(node, peer_relationships[0].name)
            if peer := await peer_relationship.get_peer(db=graphql_context.db):
                if peer.id != input_id:
                    raise ValidationError(
                        f"{node_id!r} {node.get_kind()!r} is already related to another peer on '{peer_relationships[0].name}'"
                    )


async def _collect_current_peers(
    info: GraphQLResolveInfo, data: RelationshipNodesInput, source_node: Node
) -> dict[str, RelationshipPeerData]:
    graphql_context: GraphqlContext = info.context
    relationship_name = str(data.name)

    rel_schema = source_node.get_schema().get_relationship(name=relationship_name)

    # The nodes that are already present in the db
    query = await RelationshipGetPeerQuery.init(
        db=graphql_context.db,
        source=source_node,
        rel=Relationship(schema=rel_schema, branch=graphql_context.branch, node=source_node),
    )
    await query.execute(db=graphql_context.db)
    return {str(peer.peer_id): peer for peer in query.get_peers()}


def _get_group_event_type(
    node: Node, relationship_schema: RelationshipSchema, relationship_name: str
) -> GroupUpdateType:
    group_event_type = GroupUpdateType.NONE
    if relationship_schema.identifier == "group_member":
        if "CoreGroup" in node.get_schema().inherit_from and relationship_name == "members":
            # Updating members of a group
            group_event_type = GroupUpdateType.MEMBERS

        elif relationship_name == "member_of_groups":
            # Modifying the membership of the current node
            group_event_type = GroupUpdateType.MEMBER_OF_GROUPS
    return group_event_type



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_propose_change_comment = GenericSchema(
    name="Comment",
    namespace="Core",
    description="A comment on a Proposed Change",
    label="Comment",
    display_labels=["text__value"],
    order_by=["created_at__value"],
    include_in_menu=False,
    branch=BranchSupportType.AGNOSTIC,
    attributes=[
        Attr(name="text", kind="TextArea", unique=False, optional=False),
        Attr(name="created_at", kind="DateTime", optional=True),
    ],
    relationships=[
        Rel(
            name="created_by",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=True,
            branch=BranchSupportType.AGNOSTIC,
            cardinality=Cardinality.ONE,
            identifier="comment__account",
        ),
    ],
)

core_thread = GenericSchema(
    name="Thread",
    namespace="Core",
    description="A thread on a Proposed Change",
    label="Thread",
    order_by=["created_at__value"],
    branch=BranchSupportType.AGNOSTIC,
    include_in_menu=False,
    attributes=[
        Attr(name="label", kind="Text", optional=True),
        Attr(name="resolved", kind="Boolean", default_value=False),
        Attr(name="created_at", kind="DateTime", optional=True),
    ],
    relationships=[
        Rel(
            name="change",
            peer=InfrahubKind.PROPOSEDCHANGE,
            identifier="proposedchange__thread",
            kind=RelKind.PARENT,
            optional=False,
            cardinality=Cardinality.ONE,
        ),
        Rel(
            name="comments",
            peer=InfrahubKind.THREADCOMMENT,
            identifier="thread__threadcomment",
            kind=RelKind.COMPONENT,
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
        ),
        Rel(
            name="created_by",
            peer=InfrahubKind.GENERICACCOUNT,
            identifier="thread__account",
            optional=True,
            branch=BranchSupportType.AGNOSTIC,
            cardinality=Cardinality.ONE,
        ),
    ],
)

core_change_thread = NodeSchema(
    name="ChangeThread",
    namespace="Core",
    description="A thread on proposed change",
    include_in_menu=False,
    label="Change Thread",
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.THREAD],
    generate_profile=False,
)

core_file_thread = NodeSchema(
    name="FileThread",
    namespace="Core",
    description="A thread related to a file on a proposed change",
    include_in_menu=False,
    label="Thread - File",
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.THREAD],
    generate_profile=False,
    attributes=[
        Attr(name="file", kind="Text", optional=True),
        Attr(name="commit", kind="Text", optional=True),
        Attr(name="line_number", kind="Number", optional=True),
    ],
    relationships=[
        Rel(
            name="repository",
            peer=InfrahubKind.REPOSITORY,
            optional=False,
            cardinality=Cardinality.ONE,
            branch=BranchSupportType.AGNOSTIC,
        ),
    ],
)

core_artifact_thread = NodeSchema(
    name="ArtifactThread",
    namespace="Core",
    description="A thread related to an artifact on a proposed change",
    include_in_menu=False,
    label="Thread - Artifact",
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.THREAD],
    generate_profile=False,
    attributes=[
        Attr(name="artifact_id", kind="Text", optional=True),
        Attr(name="storage_id", kind="Text", optional=True),
        Attr(name="line_number", kind="Number", optional=True),
    ],
)

core_object_thread = NodeSchema(
    name="ObjectThread",
    namespace="Core",
    description="A thread related to an object on a proposed change",
    include_in_menu=False,
    label="Thread - Object",
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.THREAD],
    generate_profile=False,
    attributes=[
        Attr(name="object_path", kind="Text", optional=False),
    ],
)

core_change_comment = NodeSchema(
    name="ChangeComment",
    namespace="Core",
    description="A comment on proposed change",
    include_in_menu=False,
    label="Change Comment",
    default_filter="text__value",
    display_labels=["text__value"],
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.COMMENT],
    generate_profile=False,
    relationships=[
        Rel(
            name="change",
            kind=RelKind.PARENT,
            peer=InfrahubKind.PROPOSEDCHANGE,
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)

core_thread_comment = NodeSchema(
    name="ThreadComment",
    namespace="Core",
    description="A comment on thread within a Proposed Change",
    include_in_menu=False,
    label="Thread Comment",
    default_filter="text__value",
    display_labels=["text__value"],
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.COMMENT],
    generate_profile=False,
    attributes=[],
    relationships=[
        Rel(
            name="thread",
            peer=InfrahubKind.THREAD,
            kind=RelKind.PARENT,
            identifier="thread__threadcomment",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)



from typing import Any, Generator

from neo4j.graph import Node as Neo4jNode

from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId


class EnrichedDiffRootsMetadataQuery(Query):
    name = "enriched_diff_roots_metadata"
    type = QueryType.READ

    def __init__(
        self,
        diff_branch_names: list[str] | None = None,
        base_branch_names: list[str] | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.diff_branch_names = diff_branch_names
        self.base_branch_names = base_branch_names
        self.from_time = from_time
        self.to_time = to_time
        self.tracking_id = tracking_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_branch_names": self.diff_branch_names,
            "base_branch_names": self.base_branch_names,
            "from_time": self.from_time.to_string() if self.from_time else None,
            "to_time": self.to_time.to_string() if self.to_time else None,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
        }

        query = """
        MATCH (diff_root:DiffRoot)
        WHERE (diff_root.is_merged IS NULL OR diff_root.is_merged <> TRUE)
        AND ($diff_branch_names IS NULL OR diff_root.diff_branch IN $diff_branch_names)
        AND ($base_branch_names IS NULL OR diff_root.base_branch IN $base_branch_names)
        AND ($from_time IS NULL OR diff_root.from_time >= $from_time)
        AND ($to_time IS NULL OR diff_root.to_time <= $to_time)
        AND ($tracking_id IS NULL OR diff_root.tracking_id = $tracking_id)
        """
        self.return_labels = ["diff_root"]
        self.add_to_query(query=query)

    def get_root_nodes_metadata(self) -> Generator[Neo4jNode, None, None]:
        for result in self.get_results():
            yield result.get_node("diff_root")



from __future__ import annotations

import uuid
from typing import Any

from typing_extensions import TYPE_CHECKING

from infrahub.workers.utils import inject_context_parameter, inject_service_parameter
from infrahub.workflows.models import WorkflowDefinition, WorkflowInfo

from . import InfrahubWorkflow, Return

if TYPE_CHECKING:
    from infrahub.context import InfrahubContext
    from infrahub.services import InfrahubServices


class WorkflowLocalExecution(InfrahubWorkflow):
    service: InfrahubServices | None = None  # needed for local injections

    async def execute_workflow(
        self,
        workflow: WorkflowDefinition,
        expected_return: type[Return] | None = None,  # noqa: ARG002
        context: InfrahubContext | None = None,
        parameters: dict[str, Any] | None = None,
        tags: list[str] | None = None,  # noqa: ARG002
    ) -> Any:
        if self.service is None:
            raise ValueError("WorkflowLocalExecution.service is not initialized")

        flow_func = workflow.load_function()
        parameters = dict(parameters) if parameters is not None else {}  # avoid mutating input parameters
        inject_service_parameter(func=flow_func, parameters=parameters, service=self.service)
        inject_context_parameter(func=flow_func, parameters=parameters, context=context)

        parameters = flow_func.validate_parameters(parameters=parameters)
        return await flow_func(**parameters)

    async def submit_workflow(
        self,
        workflow: WorkflowDefinition,
        context: InfrahubContext | None = None,
        parameters: dict[str, Any] | None = None,
        tags: list[str] | None = None,  # noqa: ARG002
    ) -> WorkflowInfo:
        await self.execute_workflow(workflow=workflow, context=context, parameters=parameters)
        return WorkflowInfo(id=uuid.uuid4())



from __future__ import annotations

import enum
from enum import Enum
from typing import TYPE_CHECKING, Any

from pydantic import field_validator, model_validator

from infrahub import config
from infrahub.core.enums import generate_python_enum
from infrahub.core.query.attribute import default_attribute_query_filter
from infrahub.types import ATTRIBUTE_KIND_LABELS, ATTRIBUTE_TYPES

from .generated.attribute_schema import GeneratedAttributeSchema

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.constants import BranchSupportType
    from infrahub.core.query import QueryElement
    from infrahub.database import InfrahubDatabase


class AttributeSchema(GeneratedAttributeSchema):
    _sort_by: list[str] = ["name"]
    _enum_class: type[enum.Enum] | None = None

    @property
    def is_attribute(self) -> bool:
        return True

    @property
    def is_relationship(self) -> bool:
        return False

    @property
    def is_deprecated(self) -> bool:
        return bool(self.deprecation)

    def to_dict(self) -> dict:
        data = self.model_dump(exclude_unset=True, exclude_none=True)
        for field_name, value in data.items():
            if isinstance(value, Enum):
                data[field_name] = value.value
        return data

    @field_validator("kind")
    @classmethod
    def kind_options(cls, v: str) -> str:
        if v not in ATTRIBUTE_KIND_LABELS:
            raise ValueError(f"Only valid Attribute Kind are : {ATTRIBUTE_KIND_LABELS} ")
        return v

    @model_validator(mode="before")
    @classmethod
    def validate_dropdown_choices(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that choices are defined for a dropdown but not for other kinds."""
        if values.get("kind") != "Dropdown" and values.get("choices"):
            raise ValueError(f"Can only specify 'choices' for kind=Dropdown: {values['kind']}")

        if values.get("kind") == "Dropdown" and not values.get("choices"):
            raise ValueError("The property 'choices' is required for kind=Dropdown")

        return values

    def get_class(self) -> type[BaseAttribute]:
        return ATTRIBUTE_TYPES[self.kind].get_infrahub_class()

    @property
    def uses_enum_class(self) -> bool:
        return bool(self.enum) and config.SETTINGS.experimental_features.graphql_enums

    def get_branch(self) -> BranchSupportType:
        if not self.branch:
            raise ValueError("branch hasn't been defined yet")
        return self.branch

    def get_enum_class(self) -> type[enum.Enum]:
        if not self.enum:
            raise ValueError(f"{self.name} is not an Enum")
        if not self._enum_class:
            self._enum_class = generate_python_enum(name=f"{self.name.title()}Enum", options=self.enum)
        return self._enum_class

    def convert_value_to_enum(self, value: Any) -> enum.Enum | None:
        if isinstance(value, enum.Enum) or value is None:
            return value
        enum_class = self.get_enum_class()
        return enum_class(value)

    def convert_enum_to_value(self, data: Any) -> Any:
        if isinstance(data, list):
            value = [self.convert_enum_to_value(element) for element in data]
            return [element.value if isinstance(element, enum.Enum) else element for element in value]
        if isinstance(data, enum.Enum):
            return data.value
        return data

    def update_from_generic(self, other: AttributeSchema) -> None:
        fields_to_exclude = ("id", "order_weight", "branch", "inherited")
        for name in self.model_fields:
            if name in fields_to_exclude:
                continue
            if getattr(self, name) != getattr(other, name):
                setattr(self, name, getattr(other, name))

    def to_node(self) -> dict[str, Any]:
        fields_to_exclude = {"id", "state", "filters"}
        fields_to_json = {"computed_attribute"}
        data = self.model_dump(exclude=fields_to_exclude | fields_to_json)

        for field_name in fields_to_json:
            if field := getattr(self, field_name):
                data[field_name] = {"value": field.model_dump()}
            else:
                data[field_name] = None

        return data

    async def get_query_filter(
        self,
        name: str,
        filter_name: str,
        branch: Branch | None = None,
        filter_value: str | int | bool | list | None = None,
        include_match: bool = True,
        param_prefix: str | None = None,
        db: InfrahubDatabase | None = None,
        partial_match: bool = False,
        support_profiles: bool = False,
    ) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
        if self.enum:
            filter_value = self.convert_enum_to_value(filter_value)

        return await default_attribute_query_filter(
            name=name,
            attribute_kind=self.kind,
            filter_name=filter_name,
            branch=branch,
            filter_value=filter_value,
            include_match=include_match,
            param_prefix=param_prefix,
            db=db,
            partial_match=partial_match,
            support_profiles=support_profiles,
        )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from prefect import task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreGenericRepository, CoreGraphQLQuery
from infrahub.core.registry import registry
from infrahub.database import InfrahubDatabase  # noqa: TC001  needed for prefect flow
from infrahub.git.utils import get_repositories_commit_per_branch
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import prepare_graphql_params

from .models import (
    ComputedAttrJinja2TriggerDefinition,
    ComputedAttrPythonQueryTriggerDefinition,
    ComputedAttrPythonTriggerDefinition,
    PythonTransformComputedAttribute,
)

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreTransformPython as CoreTransformPythonNode
    from infrahub.git.models import RepositoryData


@task(
    name="gather-python-transform-attributes",
    task_run_name="Gather Python transform attributes for {branch_name}",
    cache_policy=NONE,
)
async def gather_python_transform_attributes(
    db: InfrahubDatabase, branch_name: str, repositories: dict[str, RepositoryData] | None = None
) -> list[PythonTransformComputedAttribute]:
    log = get_run_logger()
    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    branches_with_diff_from_main = registry.get_altered_schema_branches()
    branch = registry.get_branch_from_registry(branch=branch_name)

    transform_attributes = schema_branch.computed_attributes.python_attributes_by_transform

    transform_names = list(transform_attributes.keys())

    if not transform_names:
        return []

    transforms: list[CoreTransformPythonNode] = await NodeManager.query(
        db=db,
        schema=InfrahubKind.TRANSFORMPYTHON,
        branch=branch_name,
        fields={"id": None, "name": None, "repository": None, "query": None},
        filters={"name__values": transform_names},
        prefetch_relationships=True,
    )

    found_transforms_names = [transform.name.value for transform in transforms]
    for transform_name in transform_names:
        if transform_name not in found_transforms_names:
            log.warning(
                msg=f"The transform {transform_name} is assigned to a computed attribute but the transform could not be found in the database."
            )
    repositories = repositories or await get_repositories_commit_per_branch(db=db)
    graphql_params = await prepare_graphql_params(db=db, branch=branch)

    computed_attributes: list[PythonTransformComputedAttribute] = []
    for transform in transforms:
        repository = await transform.repository.get_peer(db=db, peer_type=CoreGenericRepository, raise_on_error=True)
        query = await transform.query.get_peer(db=db, peer_type=CoreGraphQLQuery, raise_on_error=True)
        query_analyzer = InfrahubGraphQLQueryAnalyzer(
            query=query.query.value,
            branch=branch,
            schema_branch=schema_branch,
            schema=graphql_params.schema,
        )
        for attribute in transform_attributes[transform.name.value]:
            python_transform_computed_attribute = PythonTransformComputedAttribute(
                name=transform.name.value,
                branch_name=branch_name,
                repository_id=repository.get_id(),
                repository_name=repository.name.value,
                repository_kind=repository.get_kind(),
                query_analyzer=query_analyzer,
                query_name=query.name.value,
                computed_attribute=attribute,
                default_schema=branch_name not in branches_with_diff_from_main,
            )
            python_transform_computed_attribute.populate_branch_commit(
                repository_data=repositories.get(repository.name.value)
            )
            computed_attributes.append(python_transform_computed_attribute)

    return computed_attributes


@task(
    name="gather-trigger-computed-attribute-jinja2",
    cache_policy=NONE,
)
async def gather_trigger_computed_attribute_jinja2() -> list[ComputedAttrJinja2TriggerDefinition]:
    log = get_run_logger()

    # Build a list of all branches to process based on which branch is different from main
    branches_with_diff_from_main = registry.get_altered_schema_branches()
    branches_to_process: list[tuple[str, list[str]]] = [(branch, []) for branch in branches_with_diff_from_main]
    branches_to_process.append((registry.default_branch, branches_with_diff_from_main))

    triggers: list[ComputedAttrJinja2TriggerDefinition] = []

    for branch_scope, branches_out_of_scope in branches_to_process:
        schema_branch = registry.schema.get_schema_branch(name=branch_scope)
        mapping = schema_branch.computed_attributes.get_jinja2_trigger_nodes()

        log.info(f"Generating {len(mapping)} Jinja2 trigger for {branch_scope} (except {branches_out_of_scope})")

        for computed_attribute, trigger_nodes in mapping.items():
            for trigger_node in trigger_nodes:
                trigger = ComputedAttrJinja2TriggerDefinition.from_computed_attribute(
                    branch=branch_scope,
                    computed_attribute=computed_attribute,
                    trigger_node=trigger_node,
                    branches_out_of_scope=branches_out_of_scope,
                )
                triggers.append(trigger)

    return triggers


@task(
    name="gather-trigger-computed-attribute-python",
    cache_policy=NONE,
)
async def gather_trigger_computed_attribute_python(
    db: InfrahubDatabase,
) -> tuple[list[ComputedAttrPythonTriggerDefinition], list[ComputedAttrPythonQueryTriggerDefinition]]:
    triggers_python = []
    triggers_python_query = []

    repositories = await get_repositories_commit_per_branch(db=db)

    all_computed_attributes: dict[str, dict[str, PythonTransformComputedAttribute]] = defaultdict(dict)
    for branch in list(registry.branch.values()):
        if branch.is_global:
            continue

        computed_attributes = await gather_python_transform_attributes(
            db=db, branch_name=branch.name, repositories=repositories
        )
        for computed_attribute in computed_attributes:
            all_computed_attributes[computed_attribute.name][branch.name] = computed_attribute

    for branches in all_computed_attributes.values():
        branches_with_diff_from_main = []
        if registry.default_branch in branches.keys():
            commit_main = branches[registry.default_branch].repository_commit
            branches_with_diff_from_main = [
                branch_name for branch_name, item in branches.items() if item.repository_commit != commit_main
            ]
        else:
            branches_with_diff_from_main = list(branches.keys())

        branches_to_process: list[tuple[str, list[str]]] = [(branch, []) for branch in branches_with_diff_from_main]
        branches_to_process.append((registry.default_branch, branches_with_diff_from_main))

        for branch_scope, branches_out_of_scope in branches_to_process:
            trigger_python = ComputedAttrPythonTriggerDefinition.from_object(
                computed_attribute=branches[branch_scope],
                branch=branch_scope,
                branches_out_of_scope=branches_out_of_scope,
            )
            triggers_python.append(trigger_python)

            for kind in branches[branch_scope].query_analyzer.query_report.requested_read.keys():
                trigger_python_query = ComputedAttrPythonQueryTriggerDefinition.from_object(
                    kind=kind,
                    computed_attribute=branches[branch_scope],
                    branch=branch_scope,
                    branches_out_of_scope=branches_out_of_scope,
                )
                triggers_python_query.append(trigger_python_query)

    return triggers_python, triggers_python_query



from typing import Any, Iterable

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffs,
    EnrichedDiffSingleRelationship,
    EnrichedDiffsMetadata,
    EnrichedNodeCreateRequest,
)


class EnrichedDiffRootsUpsertQuery(Query):
    name = "enriched_roots_create"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, enriched_diffs: EnrichedDiffs | EnrichedDiffsMetadata, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enriched_diffs = enriched_diffs

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = self._build_diff_root_params(enriched_diffs=self.enriched_diffs)
        query = """
UNWIND $diff_root_list AS diff_root_map
WITH diff_root_map
CALL {
    WITH diff_root_map
    MERGE (diff_root:DiffRoot {uuid: diff_root_map.uuid})
    SET diff_root.base_branch = diff_root_map.base_branch
    SET diff_root.diff_branch = diff_root_map.diff_branch
    SET diff_root.from_time = diff_root_map.from_time
    SET diff_root.to_time = diff_root_map.to_time
    SET diff_root.tracking_id = diff_root_map.tracking_id
    RETURN diff_root
}
WITH DISTINCT diff_root AS diff_root
WITH collect(diff_root) AS diff_roots
WHERE SIZE(diff_roots) = 2
CALL {
    WITH diff_roots
    WITH diff_roots[0] AS base_diff_node, diff_roots[1] AS branch_diff_node
    MERGE (base_diff_node)-[:DIFF_HAS_PARTNER]-(branch_diff_node)
    SET (base_diff_node).partner_uuid = (branch_diff_node).uuid
    SET (branch_diff_node).partner_uuid = (base_diff_node).uuid
}
        """
        self.add_to_query(query)

    def _build_diff_root_params(self, enriched_diffs: EnrichedDiffs | EnrichedDiffsMetadata) -> dict[str, Any]:
        diff_root_list: list[dict[str, Any]] = []
        for enriched_diff in (enriched_diffs.base_branch_diff, enriched_diffs.diff_branch_diff):
            diff_root_list.append(
                {
                    "base_branch": enriched_diff.base_branch_name,
                    "diff_branch": enriched_diff.diff_branch_name,
                    "from_time": enriched_diff.from_time.to_string(),
                    "to_time": enriched_diff.to_time.to_string(),
                    "uuid": enriched_diff.uuid,
                    "tracking_id": enriched_diff.tracking_id.serialize() if enriched_diff.tracking_id else None,
                }
            )
        return {"diff_root_list": diff_root_list}


class EnrichedNodeBatchCreateQuery(Query):
    name = "enriched_nodes_create"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, node_create_batch: list[EnrichedNodeCreateRequest], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.node_create_batch = node_create_batch

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = self._build_node_batch_params()
        query = """
UNWIND $node_details_list AS node_details
WITH
    node_details.root_uuid AS root_uuid,
    node_details.node_map AS node_map,
    toString(node_details.node_map.node_properties.uuid) AS node_uuid
MERGE (diff_root:DiffRoot {uuid: root_uuid})
MERGE (diff_root)-[:DIFF_HAS_NODE]->(diff_node:DiffNode {uuid: node_uuid})
WITH root_uuid, node_map, diff_node, (node_map.conflict_params IS NOT NULL) AS has_node_conflict
SET
    diff_node.kind = node_map.node_properties.kind,
    diff_node.label = node_map.node_properties.label,
    diff_node.changed_at = node_map.node_properties.changed_at,
    diff_node.action = node_map.node_properties.action,
    diff_node.path_identifier = node_map.node_properties.path_identifier
WITH root_uuid, node_map, diff_node, has_node_conflict
CALL {
    // -------------------------
    // delete parent-child relationships for included nodes, they will be added in EnrichedNodesLinkQuery
    // -------------------------
    WITH diff_node
    MATCH (diff_node)-[:DIFF_HAS_RELATIONSHIP]->(:DiffRelationship)-[parent_rel:DIFF_HAS_NODE]->(:DiffNode)
    DELETE parent_rel
}
OPTIONAL MATCH (diff_node)-[:DIFF_HAS_CONFLICT]->(current_node_conflict:DiffConflict)
CALL {
    // -------------------------
    // create a node-level conflict, if necessary
    // -------------------------
    WITH diff_node, current_node_conflict, has_node_conflict
    WITH diff_node, current_node_conflict, has_node_conflict
    WHERE current_node_conflict IS NULL AND has_node_conflict = TRUE
    CREATE (diff_node)-[:DIFF_HAS_CONFLICT]->(:DiffConflict)
}
CALL {
    // -------------------------
    // delete a node-level conflict, if necessary
    // -------------------------
    WITH current_node_conflict, has_node_conflict
    WITH current_node_conflict, has_node_conflict
    WHERE current_node_conflict IS NOT NULL AND has_node_conflict = FALSE
    DETACH DELETE current_node_conflict
}
WITH root_uuid, node_map, diff_node, has_node_conflict, node_map.conflict_params AS node_conflict_params
CALL {
    // -------------------------
    // set the properties of the node-level conflict, if necessary
    // -------------------------
    WITH diff_node, has_node_conflict, node_conflict_params
    WITH diff_node, has_node_conflict, node_conflict_params
    WHERE has_node_conflict = TRUE
    OPTIONAL MATCH (diff_node)-[:DIFF_HAS_CONFLICT]->(node_conflict:DiffConflict)
    SET node_conflict = node_conflict_params
}
CALL {
    // -------------------------
    // remove stale attributes for this node
    // -------------------------
    WITH diff_node, node_map
    CALL {
        WITH diff_node, node_map
        WITH diff_node, %(attr_name_list_comp)s AS attr_names
        OPTIONAL MATCH (diff_node)-[:DIFF_HAS_ATTRIBUTE]->(attr_to_delete:DiffAttribute)
        WHERE NOT (attr_to_delete.name IN attr_names)
        OPTIONAL MATCH (attr_to_delete)-[*..6]->(next_to_delete)
        DETACH DELETE next_to_delete
        DETACH DELETE attr_to_delete
    }
    // -------------------------
    // add attributes for this node
    // -------------------------
    UNWIND node_map.attributes AS node_attribute
    MERGE (diff_node)-[:DIFF_HAS_ATTRIBUTE]->(diff_attribute:DiffAttribute {name: node_attribute.node_properties.name})
    SET diff_attribute = node_attribute.node_properties
    // -------------------------
    // add properties for this attribute
    // -------------------------
    WITH diff_attribute, node_attribute
    // -------------------------
    // remove stale properties for this attribute
    // -------------------------
    CALL {
        WITH diff_attribute, node_attribute
        WITH diff_attribute, %(attr_props_list_comp)s AS prop_types
        OPTIONAL MATCH (diff_attribute)-[:DIFF_HAS_PROPERTY]->(prop_to_delete:DiffProperty)
        WHERE NOT (prop_to_delete.property_type IN prop_types)
        OPTIONAL MATCH (prop_to_delete)-[*..4]->(next_to_delete)
        DETACH DELETE next_to_delete
        DETACH DELETE prop_to_delete
    }
    UNWIND node_attribute.properties AS attr_property
    MERGE (diff_attribute)-[:DIFF_HAS_PROPERTY]->(diff_attr_prop:DiffProperty {property_type: attr_property.node_properties.property_type})
    SET diff_attr_prop = attr_property.node_properties
    // -------------------------
    // add/remove conflict for this property
    // -------------------------
    WITH diff_attr_prop, attr_property
    OPTIONAL MATCH (diff_attr_prop)-[:DIFF_HAS_CONFLICT]->(current_attr_prop_conflict:DiffConflict)
    WITH diff_attr_prop, attr_property, current_attr_prop_conflict, (attr_property.conflict_params IS NOT NULL) AS has_prop_conflict
    FOREACH (i in CASE WHEN has_prop_conflict = FALSE THEN [1] ELSE [] END |
        DETACH DELETE current_attr_prop_conflict
    )
    FOREACH (i in CASE WHEN has_prop_conflict = TRUE THEN [1] ELSE [] END |
        MERGE (diff_attr_prop)-[:DIFF_HAS_CONFLICT]->(diff_attr_prop_conflict:DiffConflict)
        SET diff_attr_prop_conflict = attr_property.conflict_params
    )
}
// -------------------------
// remove stale relationships for this node
// -------------------------
CALL {
    WITH diff_node, node_map
    WITH diff_node, %(rel_name_list_comp)s AS rel_names
    OPTIONAL MATCH (diff_node)-[:DIFF_HAS_RELATIONSHIP]->(rel_to_delete:DiffRelationship)
    WHERE NOT (rel_to_delete.name IN rel_names)
    OPTIONAL MATCH (rel_to_delete)-[*..8]->(next_to_delete)
    DETACH DELETE next_to_delete
    DETACH DELETE rel_to_delete
}
// -------------------------
// add relationships for this node
// -------------------------
WITH diff_node, node_map
UNWIND node_map.relationships as node_relationship
MERGE (diff_node)-[:DIFF_HAS_RELATIONSHIP]->(diff_relationship:DiffRelationship {name: node_relationship.node_properties.name})
SET diff_relationship = node_relationship.node_properties
// -------------------------
// remove stale elements for this relationship group
// -------------------------
WITH diff_relationship, node_relationship
CALL {
    WITH diff_relationship, node_relationship
    WITH diff_relationship, %(rel_peers_list_comp)s AS rel_peers
    OPTIONAL MATCH (diff_relationship)-[:DIFF_HAS_ELEMENT]->(element_to_delete:DiffRelationshipElement)
    WHERE NOT (element_to_delete.peer_id IN rel_peers)
    OPTIONAL MATCH (element_to_delete)-[*..6]->(next_to_delete)
    DETACH DELETE next_to_delete
    DETACH DELETE element_to_delete
}
// -------------------------
// add elements for this relationship group
// -------------------------
WITH diff_relationship, node_relationship
UNWIND node_relationship.relationships as node_single_relationship
MERGE (diff_relationship)-[:DIFF_HAS_ELEMENT]
    ->(diff_relationship_element:DiffRelationshipElement {peer_id: node_single_relationship.node_properties.peer_id})
SET diff_relationship_element = node_single_relationship.node_properties
// -------------------------
// add/remove conflict for this relationship element
// -------------------------
WITH diff_relationship_element, node_single_relationship
OPTIONAL MATCH (diff_relationship_element)-[:DIFF_HAS_CONFLICT]->(current_element_conflict:DiffConflict)
WITH diff_relationship_element, node_single_relationship, current_element_conflict,
    (node_single_relationship.conflict_params IS NOT NULL) AS has_element_conflict
FOREACH (i in CASE WHEN has_element_conflict = FALSE THEN [1] ELSE [] END |
    DETACH DELETE current_element_conflict
)
FOREACH (i in CASE WHEN has_element_conflict = TRUE THEN [1] ELSE [] END |
    MERGE (diff_relationship_element)-[:DIFF_HAS_CONFLICT]->(element_conflict:DiffConflict)
    SET element_conflict = node_single_relationship.conflict_params
)
// -------------------------
// remove stale properties for this relationship element
// -------------------------
WITH diff_relationship_element, node_single_relationship
CALL {
    WITH diff_relationship_element, node_single_relationship
    WITH diff_relationship_element, %(element_props_list_comp)s AS element_props
    OPTIONAL MATCH (diff_relationship_element)-[:DIFF_HAS_PROPERTY]->(property_to_delete:DiffProperty)
    WHERE NOT (property_to_delete.property_type IN element_props)
    OPTIONAL MATCH (property_to_delete)-[*..4]->(next_to_delete)
    DETACH DELETE next_to_delete
    DETACH DELETE property_to_delete
}
// -------------------------
// add properties for this relationship element
// -------------------------
WITH diff_relationship_element, node_single_relationship
UNWIND node_single_relationship.properties as node_relationship_property
MERGE (diff_relationship_element)-[:DIFF_HAS_PROPERTY]
    ->(diff_relationship_property:DiffProperty {property_type: node_relationship_property.node_properties.property_type})
SET diff_relationship_property = node_relationship_property.node_properties
// -------------------------
// add conflict for this relationship element
// -------------------------
WITH diff_relationship_property, node_relationship_property
OPTIONAL MATCH (diff_relationship_property)-[:DIFF_HAS_CONFLICT]->(diff_relationship_property_conflict:DiffConflict)
WITH diff_relationship_property, node_relationship_property, diff_relationship_property_conflict,
    (node_relationship_property.conflict_params IS NOT NULL) AS has_property_conflict
FOREACH (i in CASE WHEN has_property_conflict = FALSE THEN [1] ELSE [] END |
    DETACH DELETE diff_relationship_property_conflict
)
FOREACH (i in CASE WHEN has_property_conflict = TRUE THEN [1] ELSE [] END |
    MERGE (diff_relationship_property)-[:DIFF_HAS_CONFLICT]->(property_conflict:DiffConflict)
    SET property_conflict = node_relationship_property.conflict_params
)
        """ % {
            "attr_name_list_comp": db.render_list_comprehension(
                items="node_map.attributes", item_name="node_properties.name"
            ),
            "attr_props_list_comp": db.render_list_comprehension(
                items="node_attribute.properties", item_name="node_properties.property_type"
            ),
            "rel_name_list_comp": db.render_list_comprehension(
                items="node_map.relationships", item_name="node_properties.name"
            ),
            "rel_peers_list_comp": db.render_list_comprehension(
                items="node_relationship.relationships", item_name="node_properties.peer_id"
            ),
            "element_props_list_comp": db.render_list_comprehension(
                items="node_single_relationship.properties", item_name="node_properties.property_type"
            ),
        }
        self.add_to_query(query)

    def _build_conflict_params(self, enriched_conflict: EnrichedDiffConflict) -> dict[str, Any]:
        return {
            "uuid": enriched_conflict.uuid,
            "base_branch_action": enriched_conflict.base_branch_action.value,
            "base_branch_value": enriched_conflict.base_branch_value,
            "base_branch_changed_at": enriched_conflict.base_branch_changed_at.to_string()
            if enriched_conflict.base_branch_changed_at
            else None,
            "base_branch_label": enriched_conflict.base_branch_label,
            "diff_branch_action": enriched_conflict.diff_branch_action.value,
            "diff_branch_value": enriched_conflict.diff_branch_value,
            "diff_branch_changed_at": enriched_conflict.diff_branch_changed_at.to_string()
            if enriched_conflict.diff_branch_changed_at
            else None,
            "diff_branch_label": enriched_conflict.diff_branch_label,
            "selected_branch": enriched_conflict.selected_branch.value if enriched_conflict.selected_branch else None,
            "resolvable": enriched_conflict.resolvable,
        }

    def _build_diff_property_params(self, enriched_property: EnrichedDiffProperty) -> dict[str, Any]:
        conflict_params = None
        if enriched_property.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_property.conflict)
        return {
            "node_properties": {
                "property_type": enriched_property.property_type.value,
                "changed_at": enriched_property.changed_at.to_string(),
                "previous_value": enriched_property.previous_value,
                "new_value": enriched_property.new_value,
                "previous_label": enriched_property.previous_label,
                "new_label": enriched_property.new_label,
                "action": enriched_property.action,
                "path_identifier": enriched_property.path_identifier,
            },
            "conflict_params": conflict_params,
        }

    def _build_diff_attribute_params(self, enriched_attribute: EnrichedDiffAttribute) -> dict[str, Any]:
        property_props = [
            self._build_diff_property_params(enriched_property=prop) for prop in enriched_attribute.properties
        ]
        return {
            "node_properties": {
                "name": enriched_attribute.name,
                "changed_at": enriched_attribute.changed_at.to_string(),
                "action": enriched_attribute.action.value,
                "path_identifier": enriched_attribute.path_identifier,
            },
            "properties": property_props,
        }

    def _build_diff_single_relationship_params(
        self, enriched_single_relationship: EnrichedDiffSingleRelationship
    ) -> dict[str, Any]:
        property_props = [
            self._build_diff_property_params(enriched_property=prop) for prop in enriched_single_relationship.properties
        ]
        conflict_params = None
        if enriched_single_relationship.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_single_relationship.conflict)
        return {
            "node_properties": {
                "changed_at": enriched_single_relationship.changed_at.to_string(),
                "action": enriched_single_relationship.action,
                "peer_id": enriched_single_relationship.peer_id,
                "peer_label": enriched_single_relationship.peer_label,
                "path_identifier": enriched_single_relationship.path_identifier,
            },
            "conflict_params": conflict_params,
            "properties": property_props,
        }

    def _build_diff_relationship_params(self, enriched_relationship: EnrichedDiffRelationship) -> dict[str, Any]:
        single_relationship_props = [
            self._build_diff_single_relationship_params(enriched_single_relationship=esr)
            for esr in enriched_relationship.relationships
        ]
        return {
            "node_properties": {
                "name": enriched_relationship.name,
                "identifier": enriched_relationship.identifier,
                "label": enriched_relationship.label,
                "cardinality": enriched_relationship.cardinality.value,
                "changed_at": enriched_relationship.changed_at.to_string()
                if enriched_relationship.changed_at
                else None,
                "action": enriched_relationship.action,
                "path_identifier": enriched_relationship.path_identifier,
            },
            "relationships": single_relationship_props,
        }

    def _build_diff_node_params(self, enriched_node: EnrichedDiffNode) -> dict[str, Any]:
        attribute_props = [
            self._build_diff_attribute_params(enriched_attribute=attribute) for attribute in enriched_node.attributes
        ]
        relationship_props = [
            self._build_diff_relationship_params(relationship) for relationship in enriched_node.relationships
        ]
        conflict_params = None
        if enriched_node.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_node.conflict)
        return {
            "node_properties": {
                "uuid": enriched_node.uuid,
                "kind": enriched_node.kind,
                "label": enriched_node.label,
                "changed_at": enriched_node.changed_at.to_string() if enriched_node.changed_at else None,
                "action": enriched_node.action.value,
                "path_identifier": enriched_node.path_identifier,
            },
            "conflict_params": conflict_params,
            "attributes": attribute_props,
            "relationships": relationship_props,
        }

    def _build_node_batch_params(self) -> dict[str, list[dict[str, Any]]]:
        node_details: list[dict[str, Any]] = []
        for node_create_request in self.node_create_batch:
            node_details.append(
                {
                    "root_uuid": node_create_request.root_uuid,
                    "node_map": self._build_diff_node_params(enriched_node=node_create_request.node),
                }
            )
        return {"node_details_list": node_details}


class EnrichedNodesLinkQuery(Query):
    name = "enriched_nodes_link"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, diff_root_uuid: str, diff_nodes: Iterable[EnrichedDiffNode], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.diff_root_uuid = diff_root_uuid
        self.diff_nodes = diff_nodes

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        parent_node_map: dict[str, dict[str, str]] = {}
        for diff_node in self.diff_nodes:
            if diff_node.uuid not in parent_node_map:
                parent_node_map[diff_node.uuid] = {}
            for relationship in diff_node.relationships:
                for parent_node in relationship.nodes:
                    parent_node_map[diff_node.uuid][relationship.name] = parent_node.uuid
        self.params = {"root_uuid": self.diff_root_uuid, "parent_node_map": parent_node_map}
        query = """
WITH keys($parent_node_map) AS child_node_uuids
MATCH (diff_root:DiffRoot {uuid: $root_uuid})
MATCH (diff_root)-[:DIFF_HAS_NODE]->(child_node:DiffNode)
WHERE child_node.uuid IN child_node_uuids
CALL {
    WITH diff_root, child_node
    WITH diff_root, child_node, $parent_node_map[child_node.uuid] AS sub_map
    WITH diff_root, child_node, sub_map, keys(sub_map) AS relationship_names
    MATCH (child_node)-[:DIFF_HAS_RELATIONSHIP]->(diff_rel_group:DiffRelationship)
    WHERE diff_rel_group.name IN relationship_names
    WITH diff_root, diff_rel_group, toString(sub_map[diff_rel_group.name]) AS parent_uuid
    MATCH (diff_root)-[:DIFF_HAS_NODE]->(parent_node:DiffNode {uuid: parent_uuid})
    MERGE (diff_rel_group)-[:DIFF_HAS_NODE]->(parent_node)
}
        """
        self.add_to_query(query)
